#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
A command that runs some example operations on a working nomad@FAIRDI installation
as a final integration test.
'''

import time
import os
import click

from .client import client


multi_code_example_file = 'tests/data/integration/multi_code_data.zip'
simple_example_file = 'tests/data/integration/examples_vasp.zip'


@client.command(help='Runs a few example operations as a test.')
@click.option(
    '--skip-parsers', is_flag=True,
    help='Skip extensive upload and parser tests.')
@click.option(
    '--skip-publish', is_flag=True,
    help='Skip publish the upload. Should not be done on an production environment.')
@click.option(
    '--skip-doi', is_flag=True,
    help='Skip assigning a doi to a dataset.')
@click.option(
    '--skip-mirror', is_flag=True,
    help='Skip get mirror tests.')
@click.pass_context
def integrationtests(ctx, skip_parsers, skip_publish, skip_doi, skip_mirror):
    from .client import create_client
    client = create_client()
    has_doi = False
    published = False

    print('get the upload command')
    command = client.uploads.get_upload_command().response().result.upload_command_with_name

    def get_upload(upload):
        upload = client.uploads.get_upload(
            upload_id=upload.upload_id, per_page=100).response().result

        while upload.tasks_running:
            time.sleep(0.3)
            upload = client.uploads.get_upload(
                upload_id=upload.upload_id, per_page=100).response().result

        return upload

    if not skip_parsers:
        print('upload multi code test data with curl')
        command = command.replace('<local_file>', multi_code_example_file)
        command = command.replace('<name>', 'integration_test_upload')
        command += ' -k'
        code = os.system(command)
        assert code == 0, 'curl command must be successful'
        uploads = client.uploads.get_uploads(name='integration_test_upload').response().result.results
        assert len(uploads) == 1, 'exactly one test upload must be on the server'
        upload = uploads[0]

        print('observe the upload process to be finished')
        upload = get_upload(upload)

        assert upload.tasks_status == 'SUCCESS'
        total = upload.calcs.pagination.total
        assert 100 > total > 0
        assert len(upload.calcs.results) == total

        print('delete the upload again')
        client.uploads.delete_upload(upload_id=upload.upload_id).response()
        while upload.process_running:
            upload = client.uploads.get_upload(
                upload_id=upload.upload_id).response().result

    print('upload simple data with API')
    with open(simple_example_file, 'rb') as f:
        upload = client.uploads.upload(
            name='integration test upload', file=f).response().result

    print('observe the upload process to be finished')
    upload = get_upload(upload)
    total = upload.calcs.pagination.total
    assert total > 0
    assert len(upload.calcs.results) == total

    try:
        print('get repo data')
        for calc in upload.calcs.results:
            repo = client.repo.get_repo_calc(
                upload_id=upload.upload_id, calc_id=calc.calc_id).response().result
            repo['calc_id'] == calc.calc_id

        print('get archive data')
        for calc in upload.calcs.results:
            client.archive.get_archive_calc(
                upload_id=upload.upload_id, calc_id=calc.calc_id).response()

        print('get archive logs')
        for calc in upload.calcs.results:
            client.archive.get_archive_logs(
                upload_id=upload.upload_id, calc_id=calc.calc_id).response()

        query = dict(owner='staging', upload_id=[upload.upload_id])

        print('perform repo search on data')
        search = client.repo.search(per_page=100, **query).response().result
        assert search.pagination.total >= total
        assert len(search.results) <= search.pagination.total

        print('performing archive paginated search')
        result = client.archive.post_archive_query(payload={
            'pagination': {
                'page': 1,
                'per_page': 10
            },
            'query': query
        }).response().result
        assert len(result.results) > 0

        print('performing archive scrolled search')
        result = client.archive.post_archive_query(payload={
            'scroll': {
                'scroll': True
            },
            'query': query
        }).response().result
        assert len(result.results) > 0

        print('performing download')
        client.raw.raw_files_from_query(**query)

        if not skip_publish:
            print('publish upload')
            client.uploads.exec_upload_operation(
                upload_id=upload.upload_id,
                payload=dict(operation='publish')).response()

            while upload.process_running:
                upload = client.uploads.get_upload(
                    upload_id=upload.upload_id).response().result

            assert upload.tasks_status == 'SUCCESS', 'publish must be successful'
            published = True

        print('editing upload')
        dataset = 'test_dataset'
        actions = {
            'comment': {'value': 'Test comment'},
            'references': [{'value': 'http;//test_reference.com'}],
            'coauthors': [{'value': 'author1-id'}, {'value': 'author2-id'}],
            'shared_with': [{'value': 'author3-id'}],
            'datasets': [{'value': dataset}]}

        payload = dict(actions=actions, query=dict(upload_id=[upload.upload_id]))
        result = client.repo.edit_repo(payload=payload).response().result
        assert result.success
        assert client.datasets.get_dataset(name=dataset).response().result['name'] == dataset

        print('list datasets')
        result = client.datasets.list_datasets(page=1, per_page=10).response().result
        results = result.results
        assert len(results) > 0

        if not skip_doi and published:
            print('assigning a DOI')
            result = client.datasets.assign_doi(name=dataset).response().result
            doi = result.doi
            assert doi
            has_doi = True

        if not has_doi or ctx.obj.user == 'admin':
            print('deleting dataset')
            result = client.datasets.delete_dataset(name=dataset).response().result

        if not skip_mirror and ctx.obj.user == 'admin':
            print('getting upload mirror')
            # get_upload_mirror gives 404
            payload = dict(query=dict(upload_id=upload.upload_id))
            result = client.mirror.get_uploads_mirror(payload=payload).response().result
            assert len(result) == 1
            assert len(client.mirror.get_upload_mirror(upload_id=upload.upload_id).response().result.calcs) > 0

    finally:
        if not published or ctx.obj.user == 'admin':
            print('delete the upload again')
            client.uploads.delete_upload(upload_id=upload.upload_id).response()
            while upload.process_running:
                upload = client.uploads.get_upload(
                    upload_id=upload.upload_id).response().result



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import cast, Union, Dict, Tuple, List, Any, Callable
import re

from nomad import utils
from nomad.metainfo import Definition, Section, Quantity, SubSection, Reference, QuantityReference

from .storage import ArchiveReader, ArchiveList, ArchiveObject, ArchiveError
from .query import ArchiveQueryError  # TODO


class RequiredValidationError(Exception):
    def __init__(self, msg, loc):
        super().__init__(msg)
        self.msg = msg
        self.loc = loc


class RequiredReader:
    '''
    Clients can read only the required parts of an archive. They specify the required
    parts with a required specification like this.

    .. code-block:: python
        {
            "results": "include-resolved",  // contains all the results and all its references
                                            // resolved
            "workflow": {
                "calculation_result_ref": {  // implicitly resolves the reference
                    "calculation_to_system_ref": "*"  // short-hand for include
                    "calculation_to_method_ref": "include"  // resolves and includes the target,
                                                            // but no references in the target
                                                            // are resolved.
                    "*": "*",  // includes everything else ...
                    "section_eigenvalues": "exclude"  // ... but explicitly excluded parts
                }
            }
        }

    The structure has to adheres to metainfo definitions of an archive's sub-sections and
    quantities. At each point in the specification, children can be replaced with certain
    directives.

    The different directives are:
        * include ('*' is an alias), includes whole subtree, resolves on a ref
        * include-resolved, includes whole subtree, resolves all refs recursively
        * exclude (in combination with wildcard keys), omits it from the result

    This class allows to keep a requirement specification and use it to read with it
    from given upload files and entry ids.

    Attributes:
        - required: The requirement specification as a python dictionary or directive string.
    '''

    __query_archive_key_pattern = re.compile(r'^([\s\w\-]+)(\[([-?0-9]*)(:([-?0-9]*))?\])?$')

    def __init__(self, required: Union[dict, str], root_section_def: Section = None, resolve_inplace: bool = False):
        if root_section_def is None:
            from nomad import datamodel
            self.root_section_def = datamodel.EntryArchive.m_def
        else:
            self.root_section_def = root_section_def

        self.__result_root: dict = None
        self.__archive_root: dict = None  # it is actually an ArchvieReader, but we use it as dict

        self.resolve_inplace = resolve_inplace
        self.required = self.validate(required, is_root=True)

    def __to_son(self, data):
        if isinstance(data, (ArchiveList, List)):
            data = [self.__to_son(item) for item in data]

        elif isinstance(data, ArchiveObject):
            data = data.to_dict()

        return data

    def __fix_index(self, index, length):
        if index is None:
            return index
        if index < 0:
            return max(-(length), index)
        else:
            return min(length, index)

    def validate(
            self, required: Union[str, dict], definition: Definition = None,
            loc: list = None, is_root=False) -> dict:
        '''
        Validates the required specification of this instance. It will replace all
        string directives with dicts. Those will have keys `_def` and `_directive`. It
        will add a key `_def` to all dicts. The `_def` will be the respective metainfo
        definition.

        This method will raise an exception (:class:`RequiredValidationError`) to denote
        any mismatches between the required specification and the metainfo. Also to denote
        misused directives, bad structure, etc.

        raises:
            - RequiredValidationError
        '''
        if is_root and isinstance(required, dict):
            resolve_inplace = required.get('resolve-inplace', None)
            if isinstance(resolve_inplace, bool):
                self.resolve_inplace = resolve_inplace
            elif resolve_inplace is not None:
                raise RequiredValidationError(
                    'resolve-inplace is not a bool', ['resolve-inplace'])

        if definition is None:
            definition = self.root_section_def
        if loc is None:
            loc = []

        if isinstance(definition, Quantity):
            if isinstance(definition.type, Reference):
                # TODO support quantity references
                if isinstance(definition.type, QuantityReference):
                    raise RequiredValidationError(
                        'quantity references are not supported yet', loc)
                definition = definition.type.target_section_def.m_resolved()
        elif isinstance(definition, SubSection):
            definition = definition.sub_section.m_resolved()

        is_directive = isinstance(required, str)
        if not isinstance(definition, Section) and not is_directive:
            raise RequiredValidationError(
                f'{definition.name} is not a section or reference', loc)

        if is_directive:
            # TODO support 'exclude'
            if required == 'exclude':
                raise RequiredValidationError('exclude is not supported yet', loc)
            if required not in ['*', 'include', 'exclude', 'include-resolved']:
                raise RequiredValidationError(f'{required} is not a valid directive', loc)
            return dict(_def=definition, _directive=required)

        result: Dict[str, Any] = dict(_def=definition)
        for key, value in cast(dict, required).items():
            if key == 'resolve-inplace':
                continue

            loc.append(key)
            try:
                prop, index = self.__parse_required_key(key)
            except Exception:
                raise RequiredValidationError(f'invalid key format {key}', loc)
            if prop == '*':
                # TODO support wildcards
                raise RequiredValidationError('wildcard (*) keys are not supported yet', loc)
            try:
                prop_def = cast(Section, definition).all_properties[prop]
            except KeyError:
                raise RequiredValidationError(
                    f'{definition.name} has not property {prop}', loc)
            result[key] = self.validate(value, prop_def, loc)
            result[key].update(_prop=prop, _index=index)
            loc.pop()

        return result

    def read(self, archive_reader: ArchiveReader, entry_id: str) -> dict:
        '''
        Reads the archive of the given entry id from the given archive reader and applies
        the instance's requirement specification.
        '''
        assert self.__result_root is None and self.__archive_root is None, \
            'instance cannot be used concurrently for multiple reads at the same time'

        self.__archive_root = archive_reader[utils.adjust_uuid_size(entry_id)]
        result: dict = {}
        self.__result_root = result
        result.update(**cast(dict, self.__apply_required(self.required, self.__archive_root)))

        self.__archive_root = None
        self.__result_root = None

        return result

    def __parse_required_key(self, key: str) -> Tuple[str, Union[Tuple[int, int], int]]:
        key = key.strip()
        match = RequiredReader.__query_archive_key_pattern.match(key)
        index: Union[Tuple[int, int], int] = None
        if match:
            key = match.group(1)

            # check if we have indices
            if match.group(2) is not None:
                first_index, last_index = None, None
                group = match.group(3)
                first_index = None if group == '' else int(group)

                if match.group(4) is not None:
                    group = match.group(5)
                    last_index = None if group == '' else int(group)
                    index = (0 if first_index is None else first_index, last_index)

                else:
                    index = first_index  # one item

            else:
                index = None

        else:
            raise Exception('invalid key format: %s' % key)

        return key, index

    def __resolve_refs(self, archive: dict, section_def: Section) -> dict:
        ''' Resolves all references in archive. '''
        archive = self.__to_son(archive)
        result = {}
        for prop in archive:
            value = archive[prop]
            prop_def = section_def.all_properties[prop]
            handle_item: Callable[[Any], Any] = None
            if isinstance(prop_def, SubSection):
                handle_item = lambda value: self.__resolve_refs(value, prop_def.sub_section.m_resolved())
            elif isinstance(prop_def.type, Reference) and not isinstance(prop_def.type, QuantityReference):
                target_section_def = prop_def.type.target_section_def.m_resolved()
                required = dict(_directive='include-resolved', _def=target_section_def)
                handle_item = lambda value: self.__resolve_ref(required, value)
            else:
                handle_item = lambda value: value

            if isinstance(value, (list, ArchiveList)):
                result[prop] = [handle_item(item) for item in value]
            else:
                result[prop] = handle_item(value)

        return result

    def __resolve_ref(self, required: dict, path: str) -> Union[dict, str]:
        # The archive item is a reference, the required is still a dict, the references
        # This is a simplified version of the metainfo implementation (m_resolve).
        # It implements the same semantics, but does not apply checks.
        # TODO the metainfo should also provide this implementation

        # resolve to archive_item
        if not path.startswith('/'):
            # TODO support custom reference resolution, e.g. user_id based
            return path

        resolved = self.__archive_root
        path_stack = path.strip('/').split('/')
        path_stack.reverse()

        try:
            while len(path_stack) > 0:
                prop = path_stack.pop()
                resolved = resolved[prop]
                if path_stack[-1].isdigit():
                    resolved = resolved[int(path_stack.pop())]

        except Exception:
            raise ArchiveError('could not resolve reference')

        # apply required to resolved archive_item
        resolved_result = self.__apply_required(required, resolved)

        # return or add to root depending on self.resolve_inplace
        if self.resolve_inplace:
            return resolved_result

        def _setdefault(target: Union[dict, list], prop, value_type: type):
            if isinstance(target, list):
                if target[prop] is None:
                    target[prop] = value_type()
                return target[prop]

            if prop not in target:
                target[prop] = value_type()

            return target[prop]

        path_stack = path.strip('/').split('/')
        path_stack.reverse()
        target_container: Union[dict, list] = self.__result_root
        prop_or_index: Union[str, int] = None
        while len(path_stack) > 0:
            if prop_or_index is not None:
                target_container = _setdefault(target_container, prop_or_index, dict)

            prop_or_index = path_stack.pop()
            if path_stack[-1].isdigit():
                target_list = _setdefault(target_container, prop_or_index, list)
                index = int(path_stack.pop())
                for _ in range(len(target_list), index + 1):
                    target_list.append(None)
                target_container = target_list
                prop_or_index = index

        target_container[prop_or_index] = resolved_result  # type: ignore
        return path

    def __apply_required(self, required: dict, archive_item: Union[dict, str]) -> Union[Dict, str]:
        if archive_item is None:
            return None

        directive = required.get('_directive')
        if directive is not None:
            if directive == 'include-resolved':
                if isinstance(archive_item, str):
                    return self.__resolve_ref(required, archive_item)
                return self.__resolve_refs(archive_item, required['_def'])

            elif directive in ['*', 'include']:
                return self.__to_son(archive_item)

            else:
                raise ArchiveQueryError(f'unknown directive {required}')

        if isinstance(archive_item, str):
            return self.__resolve_ref(required, archive_item)

        result: dict = {}
        for key, val in required.items():
            if key.startswith('_'):
                continue

            prop, index = val['_prop'], val['_index']

            try:
                archive_child = archive_item[prop]
                is_list = isinstance(archive_child, (ArchiveList, list))

                if index is None and is_list:
                    index = (0, None)

                elif index is not None and not is_list:
                    raise ArchiveQueryError('cannot use list key on none list %s' % prop)

                if index is None:
                    pass
                else:
                    length = len(archive_child)
                    if isinstance(index, tuple):
                        index = (self.__fix_index(index[0], length), self.__fix_index(index[1], length))
                        if index[0] == index[1]:
                            archive_child = [archive_child[index[0]]]
                        else:
                            archive_child = archive_child[index[0]: index[1]]
                    else:
                        archive_child = [archive_child[self.__fix_index(index, length)]]

                if isinstance(archive_child, (ArchiveList, list)):
                    result[prop] = [
                        self.__apply_required(val, item)
                        for item in archive_child]
                else:
                    result[prop] = self.__apply_required(val, archive_child)

            except (KeyError, IndexError):
                continue

        return result



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, Any, List
import os
import logging
import numpy as np
import re
from ase.data import chemical_symbols

from nomad.datamodel import EntryArchive
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, Method, System, SingleConfigurationCalculation


class BasicParser(FairdiParser):
    '''
    Defines a fairdi parser that parse basic quantities for sections method, system and
    single_configuration_calculation.

    Arguments:
        specifications: dictionary that will be passed on to  FairdiParser
        units_mapping: dictionary of nomad units for basic quantities such as length
        auxiliary_files: re pattern to match auxilliary files from mainfile. If no files
            are found will match files in working directory.
        kwargs: metainfo_key: re pattern pairs used to parse quantity
    '''
    def __init__(self, specifications: Dict[str, Any], **kwargs):
        super().__init__(**specifications)
        self.specifications = specifications
        self.units_mapping = kwargs.get('units_mapping', {})
        self.auxilliary_files = kwargs.get('auxilliary_files', '')
        self.mainfile_parser = TextParser()
        for key, pattern in kwargs.items():
            if isinstance(pattern, str):
                self.mainfile_parser._quantities.append(
                    Quantity(key, pattern, repeats=True, flatten=False))
            elif isinstance(pattern, tuple) and isinstance(pattern[0], str):
                self.mainfile_parser._quantities.append(
                    Quantity(key, pattern[0], str_operation=pattern[1], repeats=True))
        self._re_float = r'\-*\d+\.\d+E*e*\-*\+*\d*'
        self.auxilliary_parsers: List[TextParser] = []

    def init_parser(self):
        '''
        Initializes the mainfile and auxiliary parsers.
        '''
        self.mainfile_parser.mainfile = self.mainfile
        self.mainfile_parser.logger = self.logger

        auxilliary_files = self.mainfile_parser.get('auxilliary_files', os.listdir(self.maindir))
        # remove duplicates, maintain order
        auxilliary_files = [f for n, f in enumerate(auxilliary_files) if f not in auxilliary_files[:n]]
        self.auxilliary_parsers = []
        for filename in auxilliary_files:
            filename = os.path.basename(filename)
            if self.mainfile_parser.get('auxilliary_files') is None:
                if not self.auxilliary_files or not re.match(self.auxilliary_files, filename):
                    continue
            filename = os.path.join(self.maindir, filename)
            if not os.path.isfile(filename):
                continue
            parser = self.mainfile_parser.copy()
            parser.mainfile = filename
            parser.logger = self.logger
            self.auxilliary_parsers.append(parser)

    def parse(self, mainfile: str, archive: EntryArchive, logger=None) -> None:
        '''
        Triggers parsing of mainfile and writing parsed quantities to archive.
        '''
        self.mainfile = os.path.abspath(mainfile)
        self.maindir = os.path.dirname(self.mainfile)
        self.archive = archive
        self.logger = logger if logger is not None else logging

        self.init_parser()

        def set_value(section, key, value, unit=None, shape=None, dtype=None):
            dtype = dtype if dtype is not None else type(value)
            if value is None:
                return
            try:
                if not hasattr(value, 'units'):
                    value = np.reshape(np.array(
                        value, dtype=np.dtype(dtype)), shape) if shape is not None else dtype(value)
                    value = value * unit if unit is not None else value
                setattr(section, key, value)
            except Exception:
                pass

        def get_value(source, pattern, key=None):
            if isinstance(source, str):
                val = re.findall(pattern, source)
                return val[0] if len(val) == 1 else val
            elif isinstance(source, list):
                return [get_value(s, pattern) for s in source]
            elif isinstance(source, dict):
                return source.get(key)
            else:
                return source

        def remove_empty_section(sections, definition):
            for n in range(len(sections) - 1, -1, -1):
                empty = True
                for _ in sections[n].m_traverse():
                    empty = False
                    break
                if empty:
                    sec_run.m_remove_sub_section(definition, n)

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = self.specifications.get('code_name', '')

        energy_unit = self.units_mapping.get('energy', 1.0)
        length_unit = self.units_mapping.get('length', 1.0)
        mass_unit = self.units_mapping.get('mass')
        time_unit = self.units_mapping.get('time')

        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        for key, values in self.mainfile_parser.items():
            if values is None:
                # get if from auxiliary files
                values = []
                for parser in self.auxilliary_parsers:
                    values.extend(parser.get(key, []))
            if values is None or len(values) == 0:
                continue
            # set header quantities
            set_value(sec_run, key, values[0])
            for n, value in enumerate(values):
                # method related quantities
                if len(sec_run.section_method) <= n:
                    sec_run.m_create(Method)
                sec_method = sec_run.section_method[n]
                set_value(sec_method, key, value)

                # system related quantities
                if len(sec_run.section_system) <= n:
                    sec_run.m_create(System)
                sec_system = sec_run.section_system[n]
                set_value(sec_system, key, value)

                # calculation related quantities
                if len(sec_run.section_single_configuration_calculation) <= n:
                    sec_run.m_create(SingleConfigurationCalculation)
                sec_scc = sec_run.section_single_configuration_calculation[n]
                set_value(sec_scc, key, value)

                # specific quantities that need formatting
                if 'energy' in key:
                    shape = np.shape(value) if key == 'energy_reference_fermi' else None
                    set_value(sec_scc, key, value, energy_unit, shape, np.float64)

                if 'atom_forces' in key:
                    val = get_value(value, rf'.*({re_f}) +({re_f}) +({re_f}).*', 'atom_forces')
                    if mass_unit is not None and time_unit is not None:
                        unit = mass_unit * length_unit / time_unit ** 2
                    else:
                        unit = energy_unit / length_unit
                    set_value(sec_scc, 'atom_forces', val, unit, (np.size(val) // 3, 3), np.float64)

                if 'lattice_vectors' in key:
                    val = get_value(value, rf'({re_f}) +({re_f}) +({re_f}).*', 'lattice_vectors')
                    set_value(sec_system, 'lattice_vectors', val, length_unit, (3, 3), np.float64)
                    if val is not None:
                        sec_system.configuration_periodic_dimensions = [True, True, True]

                if 'atom_positions' in key:
                    sub_key = 'atom_positions_scaled' if 'atom_positions_scaled' in key else 'atom_positions'
                    val = get_value(value, rf'({re_f}) +({re_f}) +({re_f}).*', sub_key)
                    unit = length_unit
                    if sub_key == 'atom_positions_scaled':
                        try:
                            val = np.dot(np.array(val, dtype=np.dtype(np.float64)), sec_system.lattice_vectors.magnitude)
                            unit = 1.0
                        except Exception:
                            pass
                    set_value(sec_system, 'atom_positions', val, unit, (np.size(val) // 3, 3), np.float64)

                if 'atom_velocities' in key:
                    val = get_value(value, rf'({re_f}) +({re_f}) +({re_f}).*', 'atom_velocities')
                    set_value(sec_system, 'atom_velocities', val, length_unit / time_unit, (np.size(val) // 3, 3), np.float64)

                if 'atom_labels' in key:
                    val = get_value(value, r'([A-Z][a-z]*)\s', 'atom_labels')
                    val = [val] if isinstance(val, str) else val
                    set_value(sec_system, 'atom_labels', val, shape=(len(val)), dtype=str)

                if 'atom_atom_number' in key:
                    val = get_value(value, r'(\d+)\s', 'atom_atom_number')
                    val = [val] if isinstance(val, str) else val
                    set_value(sec_system, 'atom_atom_number', val, shape=(len(val)), dtype=np.int32)
                    set_value(sec_system, 'atom_labels', [chemical_symbols[int(n)] for n in sec_system.atom_atom_number], shape=(len(val)))

        # remove unfilled sections
        remove_empty_section(sec_run.section_method, Run.section_method)
        remove_empty_section(sec_run.section_system, Run.section_system)
        remove_empty_section(sec_run.section_single_configuration_calculation, Run.section_single_configuration_calculation)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.metainfo import MSection, Quantity, Package, Datetime, SubSection

m_package = Package()


class QuantumCircuit(MSection):
    processors = Quantity(type=str, shape=['0..*'])
    number_of_registers = Quantity(type=int)
    simulated = Quantity(type=bool)


class QuantumCMS(MSection):
    '''
    The root section for all (meta)data that belongs to a single calculation.
    '''
    chemical_formula = Quantity(
        type=str,
        description=''' The chemical formula that describes the simulated material ''')

    chemical_name = Quantity(
        type=str,
        description=''' The chemical name that describes the simulated material ''')

    atom_labels = Quantity(
        type=str, shape=['1..*'],
        description=''' Labels for the atoms/elements that comprise the simulated material ''')

    transformation = Quantity(type=str)
    quantum_computer_system = Quantity(type=str)
    quantum_computing_libraries = Quantity(type=str, shape=['0..*'])
    computation_datetime = Quantity(type=Datetime)

    number_of_shots = Quantity(type=int)
    quantum_volume = Quantity(type=int)

    section_quantum_circuit = SubSection(sub_section=QuantumCircuit)


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class GulpParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        super().__init__(
            specifications=dict(
                name='parsers/gulp', code_name='gulp', code_homepage='http://gulp.curtin.edu.au/gulp/',
                mainfile_contents_re=(
                    r'\s*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*'
                    r'\*\*\*\*\*\*\*\*\*\*\*\*\*\s*'
                    r'\s*\*\s*GENERAL UTILITY LATTICE PROGRAM\s*\*\s*')),
            units_mapping=dict(length=ureg.angstrom, energy=ureg.eV),
            # include code name to distinguish gamess and firefly
            program_version=r'Version *\= *([\d\.]+)',
            lattice_vectors=r'Final Cartesian lattice vectors \(Angstroms\) \:\s*([\d\.\-\s]+)',
            atom_labels_atom_positions_scaled=r'Final asymmetric unit coordinates \:\s*\-+\s*.+\s*.+\s*\-+\s*([\s\S]+?)\-{10}',
            energy_total=rf'Final energy *= *({re_f})|Total energy *\(eV\) *\= *({re_f})')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Allows to create pydantic models from section definitions.
'''

from typing import cast
from pydantic import create_model, Field, BaseConfig
from datetime import datetime

from .metainfo import DefinitionAnnotation, Definition, Section, Quantity, Datetime, MEnum, Capitalized, JSON


class _OrmConfig(BaseConfig):
    orm_mode = True


class PydanticModel(DefinitionAnnotation):
    '''
    This annotation class can be used to extend metainfo sections. It will create a
    pydantic model from the section definition. Its a SectionAnnotation and allows
    to create pydantic model instances from section instances.

    Attributes:
        model: The pydantic model that represents the section defintion.
    '''
    def __init__(self):
        self.model = None

    def to_pydantic(self, section):
        ''' Returns the pydantic model instance for the given section. '''
        return self.model.from_orm(section)

    def init_annotation(self, definition: Definition):
        section_definition = cast(Section, definition)
        name = section_definition.name

        def create_field(quantity: Quantity):
            pydantic_type: type = None
            if quantity.type == Datetime:
                pydantic_type = datetime
            elif isinstance(quantity.type, MEnum):
                pydantic_type = str
            elif quantity.type == Capitalized:
                pydantic_type = str
            elif quantity.type == JSON:
                pydantic_type = dict
            else:
                pydantic_type = quantity.type

            return pydantic_type, Field(quantity.default, description=quantity.description)

        fields = {
            name: create_field(quantity)
            for name, quantity in section_definition.all_quantities.items()
        }

        self.model = create_model(name, __config__=_OrmConfig, **fields)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The archive storage is made from two tiers. First the whole archive is stored in
files, secondly parts of the archive are stored in mongodb documents.

The file storage is done in msg-pack files. Each file contains the archive of many
entries (all entries of an upload). These msg-pack files contain the JSON serialized
version of the metainfo archive (see module:`nomad.metainfo`). In addition msg-pack
contains TOC information for quicker access of individual sections. See :func:`write_archive`
and :func:`read_archvive`. In addition there query functionality to partially read
specified sections from an archive: func:`query_archive`.

The mongo storage uses mongodb's native bson to store JSON serialized metainfo archive
data. Each document in mongodb holds the partial archive of single entry. Which parts
of an archive are stored in mongo is determined by the metainfo and
section annotations/categories.
'''

from .storage import (
    write_archive, read_archive, ArchiveError, ArchiveReader, ArchiveWriter,
    ArchiveObject, ArchiveList, ArchiveItem)
from .query import query_archive, filter_archive, ArchiveQueryError
from .partial import (
    read_partial_archive_from_mongo, read_partial_archives_from_mongo,
    write_partial_archive_to_mongo, delete_partial_archives_from_mongo,
    create_partial_archive, compute_required_with_referenced)
from .required import RequiredReader, RequiredValidationError



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='gpaw_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='gpaw.nomadmetainfo.json'))


class x_gpaw_section_paw_method(MSection):
    '''
    GPAW PAW specific information
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gpaw_section_paw_method'))

    x_gpaw_number_of_packed_ap_elements = Quantity(
        type=int,
        shape=[],
        description='''
        number of PAW projector matrix elements in packed format
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_number_of_packed_ap_elements'))

    x_gpaw_number_of_projectors = Quantity(
        type=int,
        shape=[],
        description='''
        number of PAW projectors
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_number_of_projectors'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_gpaw_atomic_density_matrices = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'x_gpaw_number_of_packed_ap_elements'],
        description='''
        atomic density matrices in the PAW formalism
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_atomic_density_matrices'))

    x_gpaw_fixed_spin_Sz = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Target value (fixed) of the z projection of the spin moment operator $S^z$ for the
        converged calculation with the XC_method.
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_fixed_spin_Sz'))

    x_gpaw_magnetic_moments = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Magnetic moments projected onto atoms. The sum gives the total magnetic moment
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_magnetic_moments'))

    x_gpaw_projections_imag = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues', 'x_gpaw_number_of_projectors'],
        description='''
        projections in the PAW formalism (imaginary part)
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_projections_imag'))

    x_gpaw_projections_real = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues', 'x_gpaw_number_of_projectors'],
        description='''
        projections in the PAW formalism (real part)
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_projections_real'))

    x_gpaw_spin_Sz = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the z projection of the spin moment operator $S^z$ for the converged
        calculation with the XC_method.
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_spin_Sz'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_gpaw_density_convergence_criterion = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Density convergence criteria for break the SCF cycle
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_density_convergence_criterion'))

    x_gpaw_fix_density = Quantity(
        type=bool,
        shape=[],
        description='''
        Was it a calculation with a fixed density?
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_fix_density'))

    x_gpaw_fix_magnetic_moment = Quantity(
        type=bool,
        shape=[],
        description='''
        Was the magnetic moment fixed? If yes the x_gpaw_fixed_sZ is set
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_fix_magnetic_moment'))

    x_gpaw_maximum_angular_momentum = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maxium angular momentum (L) for projectors
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_maximum_angular_momentum'))

    x_gpaw_mix_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Damping parameter in the density mixer
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_mix_beta'))

    x_gpaw_mix_class = Quantity(
        type=str,
        shape=[],
        description='''
        The density mixer class name (Mixer, MixerSum, MixerDiff)
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_mix_class'))

    x_gpaw_mix_old = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of old densities in the density mixer
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_mix_old'))

    x_gpaw_mix_weight = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Mixing weight in density mixer
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_mix_weight'))

    x_gpaw_symmetry_time_reversal_switch = Quantity(
        type=bool,
        shape=[],
        description='''
        Was time reserval symmetry used
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_symmetry_time_reversal_switch'))

    x_gpaw_xc_functional = Quantity(
        type=str,
        shape=[],
        description='''
        The XC functional name used in gpaw as input
        ''',
        a_legacy=LegacyDefinition(name='x_gpaw_xc_functional'))

    x_gpaw_section_paw_method = SubSection(
        sub_section=SectionProxy('x_gpaw_section_paw_method'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gpaw_section_paw_method'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class AmberParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'
        super().__init__(
            specifications=dict(
                name='parsers/amber', code_name='Amber', domain='dft',
                mainfile_contents_re=r'\s*Amber\s[0-9]+\s[A-Z]+\s*[0-9]+'),
            units_mapping=dict(length=ureg.angstrom, energy=ureg.eV),
            program_version=r'Amber\s*(\d+)\s*(\w+)\s*(\d+)',
            # will only read initial coordinates
            auxilliary_files=r'(\S+\.inpcrd|\S+\.prmtop)',
            atom_positions=(
                rf'\d+\s+({re_f} +{re_f} +{re_f} +{re_f} +{re_f} +{re_f}[\s\S]+)',
                lambda x: x.strip().split()),
            atom_atom_number=r'\%FLAG ATOMIC_NUMBER\s*\%FORMAT\(.+\)\s*([\d\s]+)',
            energy_total=rf'NSTEP\s*ENERGY.+\s*\d+\s*({re_f})')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import itertools
import math
from functools import reduce
from typing import List, Dict, Tuple, Any, Union, cast
from nptyping import NDArray

from ase.utils import pbc2pbc
import ase.geometry

import numpy as np
from scipy.spatial import Voronoi  # pylint: disable=no-name-in-module
from matid.symmetry import WyckoffSet

from nomad.aflow_prototypes import aflow_prototypes
from nomad.constants import atomic_masses


def get_summed_atomic_mass(atomic_numbers: NDArray[Any]) -> float:
    """Calculates the summed atomic mass for the given atomic numbers.

    Args:
        atomic_numbers: Array of valid atomic numbers

    Returns:
        The atomic mass in kilograms.
    """
    # It is assumed that the atomic numbers are valid at this point.
    mass = np.sum(atomic_masses[atomic_numbers])
    return mass


def get_volume(basis: NDArray[Any]) -> float:
    """Calculates the volume of the given parallelepiped.

    Args:
        basis: 3x3 matrix with basis vectors of a parallellepiped as rows.

    Returns:
        Volume of the parallelepiped defined by the basis.
    """
    return np.abs(np.linalg.det(basis))


def wrap_positions(
        positions: NDArray[Any],
        cell: NDArray[Any] = None,
        pbc: Union[bool, NDArray[Any]] = True,
        center: NDArray[Any] = [0.5, 0.5, 0.5],
        eps: float = 1e-12) -> NDArray[Any]:
    """Wraps the given position so that they are within the unit cell. If no
    cell is given, scaled positions are assumed. For wrapping cartesian
    positions you also need to provide the cell.

    Args:
        positions: Positions of the atoms. Accepts both scaled and
            cartesian positions.
        cell: Lattice vectors for wrapping cartesian positions.
        pbc: For each axis in the unit cell decides whether the positions will
            be wrapped along this axis.
        center: The position in fractional coordinates that the wrapped
            positions will be nearest possible to.
        eps: Small number to prevent slightly negative coordinates from being
            wrapped.
    """
    if not hasattr(center, '__len__'):
        center = (center,) * 3

    pbc = pbc2pbc(pbc)
    shift = np.asarray(center) - 0.5 - eps

    # Don't change coordinates when pbc is False
    shift[np.logical_not(pbc)] = 0.0

    if cell is None:
        fractional = positions
    else:
        fractional = to_scaled(positions, cell)
    fractional -= shift

    for i, periodic in enumerate(pbc):
        if periodic:
            fractional[:, i] %= 1.0
        fractional[:, i] += shift[i]
    if cell is not None:
        return np.dot(fractional, cell)
    else:
        return fractional


def to_scaled(
        positions: NDArray[Any],
        cell: NDArray[Any] = None) -> NDArray[Any]:
    """Converts cartesian positions into scaled position one using the given
    cell lattice vectors as a basis.

    Args:
        positions: Scaled positions.
        cell: Lattice vectors.

    Returns:
        The given positions in scaled coordinates.
    """
    return np.linalg.solve(complete_cell(cell).T, positions.T).T


def to_cartesian(
        positions: NDArray[Any],
        cell: NDArray[Any] = None) -> NDArray[Any]:
    """Converts scaled positions into cartesian one using the given cell
    lattice vectors as a basis.

    Args:
        positions: Scaled positions.
        cell: Lattice vectors.

    Returns:
        The given positions in cartesian coordinates.
    """
    cartesian_positions = np.dot(positions, complete_cell(cell))
    return cartesian_positions


def complete_cell(cell: NDArray[Any]) -> NDArray[Any]:
    """Creates placeholder axes for cells with zero-dimensional lattice vectors
    in order to do linear algebra.

    Args:
        cell: Lattice vectors.

    Returns:
        The given cell with zero-dimensional lattice vectors filled with
        placeholder axes.
    """
    return ase.geometry.complete_cell(cell)


def reciprocal_cell(cell: NDArray[Any]) -> NDArray[Any]:
    """Returns the reciprocal cell without the factor or 2*Pi.

    Args:
        cell: Lattice vectors.

    Returns:
        Reciprocal cell as a 3x3 array.
    """
    return np.linalg.pinv(cell).transpose()


def find_match(pos: NDArray[Any], positions: NDArray[Any], eps: float) -> Union[int, None]:
    """Attempts to find a position within a larger list of positions.

    Args:
        pos: The point to search for
        positions: The points within which the search is performed.
        eps: Match tolerance.

    Returns:
        Index of the matched position or None if match not found.
    """
    displacements = positions - pos
    distances = np.linalg.norm(displacements, axis=1)
    min_arg = np.argmin(distances)
    min_value = distances[min_arg]
    if min_value <= eps:
        return cast(int, min_arg)
    else:
        return None


def cellpar_to_cell(cellpar: NDArray[Any], ab_normal: NDArray[Any] = [0, 0, 1], a_direction: NDArray[Any] = None, degrees=False) -> NDArray[Any]:
    """Creates a 3x3 cell from the given lattice_parameters.

    The returned cell is orientated such that a and b are normal to `ab_normal`
    and a is parallel to the projection of `a_direction` in the a-b plane.

    Default `a_direction` is (1,0,0), unless this is parallel to `ab_normal`,
    in which case default `a_direction` is (0,0,1).

    The returned cell has the vectors va, vb and vc along the rows. The cell
    will be oriented such that va and vb are normal to `ab_normal` and va will
    be along the projection of `a_direction` onto the a-b plane.

    Args:
        cellpar: Six lattice parameters: [a, b, c, alpha, beta, gamma].
        The following typical convention is used:

            a = length of first basis vector
            b = length of second basis vector
            c = length of third basis vector
            alpha = angle between b and c in radians
            beta  = angle between a and c in radians
            gamma = angle between a and b in radians
        degrees: Use degrees in place of radians.

    Returns:
        Six parameters (in this order) as a numpy
        array. Here is an explanation of each parameter:
    """
    if not degrees:
        cellpar[3:6] *= 180.0 / np.pi

    return ase.geometry.cell.cellpar_to_cell(cellpar, ab_normal, a_direction)


def cell_to_cellpar(cell: NDArray[Any], degrees=False) -> NDArray[Any]:
    """Returns lattice parameters for the given cell.

    Args:
        normalized_cell: The normalized cell as a 3x3 array. Each row is a
            basis vector.
        degrees: Use degrees in place of radians.

    Returns:
        Six parameters [a, b, c, alpha, beta, gamma]. The following typical
        convention ik used:

            a = length of first basis vector
            b = length of second basis vector
            c = length of third basis vector
            alpha = angle between b and c in radians
            beta  = angle between a and c in radians
            gamma = angle between a and b in radians
    """
    # Lengths
    lengths = np.linalg.norm(cell, axis=1)

    # Angles
    angles = np.zeros(3)
    for i in range(3):
        j = (i + 1) % 3
        k = (i + 2) % 3
        angles[i] = np.dot(
            cell[j],
            cell[k]) / (lengths[j] * lengths[k])
    angles = np.arccos(np.clip(angles, -1.0, 1.0))
    if degrees:
        angles *= 180.0 / np.pi

    return np.concatenate((lengths, angles), axis=0)


def get_symmetry_string(space_group: int, wyckoff_sets: List[WyckoffSet], is_2d: bool = False) -> str:
    """Used to serialize symmetry information into a string. The Wyckoff
    positions are assumed to be normalized and ordered as is the case if using
    the matid-library.

    Args:
        space_group: 3D space group number
        wyckoff_sets: Wyckoff sets that map a Wyckoff letter to related
            information
        is_2d: Whether the symmetry information is analyzed from a 2D
            structure. If true, a prefix is added to the string to distinguish
            2D from 3D.

    Returns:
        A string that encodes the symmetry properties of an atomistic
        structure.
    """
    wyckoff_strings = []
    for group in wyckoff_sets:
        element = group.element
        wyckoff_letter = group.wyckoff_letter
        n_atoms = len(group.indices)
        i_string = "{} {} {}".format(element, wyckoff_letter, n_atoms)
        wyckoff_strings.append(i_string)
    wyckoff_string = ", ".join(sorted(wyckoff_strings))
    if is_2d:
        string = "2D {} {}".format(space_group, wyckoff_string)
    else:
        string = "{} {}".format(space_group, wyckoff_string)

    return string


def get_hill_decomposition(atom_labels: NDArray[Any], reduced: bool = False) -> Tuple[List[str], List[int]]:
    """Given a list of atomic labels, returns the chemical formula using the
    Hill system (https://en.wikipedia.org/wiki/Hill_system) with an exception
    for binary ionic compounds where the cation is always given first.

    Args:
        atom_labels: Atom labels.
        reduced: Whether to divide the number of atoms by the greatest common
            divisor

    Returns:
        An ordered list of chemical symbols and the corresponding counts.
    """
    # Count occurancy of elements
    names = []
    counts = []
    unordered_names, unordered_counts = np.unique(atom_labels, return_counts=True)
    element_count_map = dict(zip(unordered_names, unordered_counts))

    # Apply basic Hill system:
    # 1. Is Carbon part of the system?
    if "C" in element_count_map:
        names.append("C")
        counts.append(element_count_map["C"])
        del element_count_map['C']

        # 1a. add hydrogren
        if "H" in element_count_map:
            names.append("H")
            counts.append(element_count_map["H"])
            del element_count_map["H"]

    # 2. all remaining elements in alphabetic order
    for element in sorted(element_count_map):
        names.append(element)
        counts.append(element_count_map[element])

    # 3. Binary ionic compounds: cation first, anion second
    # If any of the most electronegative elements is first
    # by alphabetic order, we move it to second
    if len(counts) == 2 and names != ["C", "H"]:
        order = {
            "F": 1,
            "O": 2,
            "N": 3,
            "Cl": 4,
            "Br": 5,
            "C": 6,
            "Se": 7,
            "S": 8,
            "I": 9,
            "As": 10,
            "H": 11,
            "P": 12,
            "Ge": 13,
            "Te": 14,
            "B": 15,
            "Sb": 16,
            "Po": 17,
            "Si": 18,
            "Bi": 19
        }
        if (names[0] in order):
            if (names[1] in order):
                if(order[names[0]] < order[names[1]]):
                    # For non-metals:
                    # Swap symbols and counts if first element
                    # is more electronegative than the second one,
                    # because the more electronegative element is the anion
                    names[0], names[1] = names[1], names[0]
                    counts[0], counts[1] = counts[1], counts[0]
            else:
                # Swap symbols and counts always if second element
                # is any other element,i.e.,
                # put non-metal last because it is the anion
                names[0], names[1] = names[1], names[0]
                counts[0], counts[1] = counts[1], counts[0]

    # TODO: implement all further exceptions regarding ordering
    #       in chemical formulas:
    #         - ionic compounds (ordering wrt to ionization)
    #         - oxides, acids, hydroxides...

    # Reduce if requested
    if reduced:
        greatest_common_divisor = reduce(math.gcd, counts)
        counts = np.array(counts) / greatest_common_divisor

    return names, counts


def get_formula_string(symbols: List[str], counts: List[int]) -> str:
    """Used to form a single formula string from a list of chemical species and
    their counts.

    Args:
        symbols: List of chemical species
        counts: List of chemical species occurences

    Returns:
        The formula as a string.
    """
    formula = ""
    for symbol, count in zip(symbols, counts):
        if count > 1:
            formula += "%s%d" % (symbol, count)
        else:
            formula += symbol
    return formula


def get_normalized_wyckoff(atomic_numbers: NDArray[Any], wyckoff_letters: NDArray[Any]) -> Dict[str, Dict[str, int]]:
    """Returns a normalized Wyckoff sequence for the given atomic numbers and
    corresponding wyckoff letters. In a normalized sequence the chemical
    species are "anonymized" by replacing them with upper case alphabets.

    Args:
        atomic_numbers: Array of atomic numbers.
        wyckoff_letters: Array of Wyckoff letters as strings.

    Returns:
        Returns a dictionary that maps each present Wyckoff letter to a
        dictionary. The dictionary contains the number of atoms for each
        species, where the species names have been anomymized in the form
        "X_<index>".
    """
    # Count the occurrence of each chemical species
    atom_count: Dict[int, int] = {}
    for atomic_number in atomic_numbers:
        atom_count[atomic_number] = atom_count.get(atomic_number, 0) + 1

    # Form a dictionary that maps Wyckoff letters to a dictionary with the
    # number of atoms of that Wyckoff letter for each atomic number
    wyc_dict: dict = {}
    for i, wyckoff_letter in enumerate(wyckoff_letters):
        old_val = wyc_dict.get(wyckoff_letter, {})
        atomic_number = atomic_numbers[i]
        old_val[atomic_number] = old_val.get(atomic_number, 0) + 1
        wyc_dict[wyckoff_letter] = old_val
    sorted_wyckoff_letters = list(wyc_dict.keys())
    sorted_wyckoff_letters.sort()

    # Anonymize the atomic species to X_<index>, where the index is calculated
    # by ordering the species.
    def compare_atomic_number(at1, at2):
        def cmpp(a, b):
            return ((a < b) - (a > b))

        c = cmpp(atom_count[at1], atom_count[at2])
        if (c != 0):
            return c
        for wyckoff_letter in sorted_wyckoff_letters:
            p = wyc_dict[wyckoff_letter]
            c = cmpp(p.get(at1, 0), p.get(at2, 0))
            if c != 0:
                return c
        return 0
    sorted_species = list(atom_count.keys())
    sorted_species.sort(key=functools.cmp_to_key(compare_atomic_number))
    standard_atom_names = {}
    for i, at in enumerate(sorted_species):
        standard_atom_names[at] = ("X_%d" % i)

    # Rename with anonymized species labels
    standard_wyc: dict = {}
    for wk, ats in wyc_dict.items():
        std_ats = {}
        for at, count in ats.items():
            std_ats[standard_atom_names[at]] = count
        standard_wyc[wk] = std_ats

    # Divide atom counts with greatest common divisor
    if standard_wyc:
        counts = [c for x in standard_wyc.values() for c in x.values()]
        gcd = counts[0]
        for c in counts[1:]:
            gcd = math.gcd(gcd, c)
        if gcd != 1:
            for d in standard_wyc.values():
                for at, c in d.items():
                    d[at] = c // gcd

    return standard_wyc


def search_aflow_prototype(space_group: int, norm_wyckoff: dict) -> dict:
    """Searches the AFLOW prototype library for a match for the given space
    group and normalized Wyckoff sequence. The normalized Wyckoff sequence is
    assumed to come from the MatID symmetry routine.

    Currently only contains Part I of the prototype library (M. J. Mehl, D.
    Hicks, C. Toher, O. Levy, R. M. Hanson, G. L. W. Hart, and S. Curtarolo,
    The AFLOW Library of Crystallographic Prototypes: Part 1, Comp. Mat. Sci.
    136, S1-S828 (2017), 10.1016/j.commatsci.2017.01.017)

    Args:
        space_group_number: Space group number
        norm_wyckoff: Normalized Wyckoff occupations

    Returns:
        Dictionary containing the AFLOW prototype information.
    """
    structure_type_info = None
    type_descriptions: Any = aflow_prototypes["prototypes_by_spacegroup"].get(space_group, [])
    for type_description in type_descriptions:
        current_norm_wyckoffs = type_description.get("normalized_wyckoff_matid")
        if current_norm_wyckoffs and current_norm_wyckoffs == norm_wyckoff:
            structure_type_info = type_description
            break
    return structure_type_info


def get_brillouin_zone(reciprocal_lattice: NDArray[Any]) -> dict:
    """Calculates the Brillouin Zone information from the given reciprocal
    lattice.

    This function uses the crystallographic definition, so there is no factor
    of 2*Pi.

    Args:
        primitive_lattice: The primitive cell as a matrix where rows are the
            cell basis vectors.

    Returns:
        A dictionary containing:
        "vertices": The vertices of the first Brillouin zone
        "faces": The indices of the vertices that make up the faces on the
            first Brillouin zone. The order of these indices matter, because
            only when combined sequentially they form the correct face.
    """
    # Create the near lattice points that surround the origin
    b1 = reciprocal_lattice[0, :]
    b2 = reciprocal_lattice[1, :]
    b3 = reciprocal_lattice[2, :]

    list_k_points = []
    for i, j, k in itertools.product([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]):
        list_k_points.append(i * b1 + j * b2 + k * b3)

    # Create the first Brillouin zone by calculating a Voronoi cell starting
    # from the reciprocal cell origin.
    voronoi = Voronoi(list_k_points)
    origin_index = 13

    # Get the vertices. The regions attribute will contain a list of
    # different regions that were found during the Voronoi creation. We want
    # the Voronoi region for the point at the origin.
    point_region = voronoi.point_region[13]
    vertice_indices = voronoi.regions[point_region]
    vertices = voronoi.vertices[vertice_indices].tolist()

    # Create a mapping between the original index and an index in the new list
    index_map = {
        old_id: new_id for (new_id, old_id) in enumerate(vertice_indices)
    }

    # The ridges are the faces of a 3D Voronoi cell. Here we search for ridges
    # that are placed between the origin and some other point. These form the
    # BZ faces.
    faces = []
    for key in voronoi.ridge_dict:
        if key[0] == origin_index or key[1] == origin_index:
            ridge_indices = voronoi.ridge_dict[key]
            new_ridge_indices = [index_map[i] for i in ridge_indices]
            faces.append(new_ridge_indices)
    faces = faces

    brillouin_zone = {
        "vertices": vertices,
        "faces": faces,
    }
    return brillouin_zone



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import logging
import json
from datetime import datetime
from ase.spacegroup import crystal as asecrystal
import numpy as np

from nomad.parsing import FairdiParser

from nomad.datamodel.metainfo.common_dft import Run, SingleConfigurationCalculation, System


class OpenKIMParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/openkim', code_name='OpenKIM', domain='dft',
            mainfile_contents_re=r'OPENKIM')

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging.getLogger('__name__')

        with open(self.filepath) as f:
            self.json = json.load(f)

        def set_value(section, key, val, unit=None):
            if val is None:
                return
            val = val * unit if unit is not None else val
            setattr(section, key, val)

        def get_value_list(entry, key):
            val = entry.get(key, [])
            return val if isinstance(val, list) else [val]

        def get_crystal(entry):
            symbols = entry.get('species.source-value', [])
            basis = entry.get('basis-atom-coordinates.source-value', [])
            spacegroup = entry.get('space-group.source-value', 1)
            cellpar_a = entry.get('a.si-value', 1)
            cellpar_b = entry.get('a.si-value', cellpar_a)
            cellpar_c = entry.get('b.si-value', cellpar_a)
            # TODO are angles denoted by alpha, beta, gamma in openkim? can they be lists?
            alpha = entry.get('alpha.source-value', 90)
            beta = entry.get('beta.source-value', 90)
            gamma = entry.get('gamma.source-value', 90)

            if isinstance(cellpar_a, float):
                cellpar_a, cellpar_b, cellpar_c = [cellpar_a], [cellpar_b], [cellpar_c]

            atoms = []
            for n in range(len(cellpar_a)):
                try:
                    atoms.append(asecrystal(
                        symbols=symbols, basis=basis, spacegroup=spacegroup, cellpar=[
                            cellpar_a[n], cellpar_b[n], cellpar_c[n], alpha, beta, gamma]))
                except Exception:
                    pass
            return atoms

        for entry in self.json:
            sec_run = self.archive.m_create(Run)
            sec_run.program_name = 'OpenKIM'
            set_value(sec_run, 'program_version', entry.get('meta.runner.short-id'))

            compile_date = entry.get('meta.created_on')
            if compile_date is not None:
                dt = datetime.strptime(compile_date, '%Y-%m-%d %H:%M:%S.%f') - datetime(1970, 1, 1)
                sec_run.program_compilation_datetime = dt.total_seconds()

            crystals = get_crystal(entry)
            for crystal in crystals:
                sec_system = sec_run.m_create(System)
                sec_system.atom_labels = crystal.get_chemical_symbols()
                sec_system.atom_positions = crystal.get_positions()
                sec_system.lattice_vectors = crystal.get_cell().array
                sec_system.configuration_periodic_dimensions = [True, True, True]

            energies = get_value_list(entry, 'cohesive-potential-energy.si-value')
            temperatures = get_value_list(entry, 'temperature.si-value')
            for n, energy in enumerate(energies):
                sec_scc = sec_run.m_create(SingleConfigurationCalculation)
                sec_scc.energy_total = energy
                if temperatures:
                    sec_scc.temperature = temperatures[n]

            stress = entry.get('cauchy-stress.si-value')
            if stress is not None:
                stress_tensor = np.zeros((3, 3))
                stress_tensor[0][0] = stress[0]
                stress_tensor[1][1] = stress[1]
                stress_tensor[2][2] = stress[2]
                stress_tensor[1][2] = stress[2][1] = stress[3]
                stress_tensor[0][2] = stress[2][0] = stress[4]
                stress_tensor[0][1] = stress[1][0] = stress[5]

        # TODO implement openkim specific metainfo



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Quantum computational materials science metadata
'''

from nomad import config
from nomad.metainfo import Quantity, MSection, Section, Datetime
from nomad.metainfo.search_extension import Search


class QCMSMetadata(MSection):
    m_def = Section(a_domain='qcms')

    # sample quantities
    chemical = Quantity(type=str, default='not processed', a_search=Search())
    quantum_computer_system = Quantity(type=str, a_search=Search())
    quantum_computing_libraries = Quantity(type=str, shape=['0..*'], a_search=Search())
    computation_datetime = Quantity(type=Datetime, a_search=Search())

    # TODO move
    quantities = Quantity(type=str, shape=['0..*'], default=[], a_search=Search())

    def apply_domain_metadata(self, entry_archive):
        if entry_archive is None:
            return

        entry = self.m_parent

        root_section = entry_archive.section_quantum_cms
        entry.formula = root_section.chemical_formula
        if not entry.formula:
            entry.formula = config.services.unavailable_value
        atoms = root_section.atom_labels
        if hasattr(atoms, 'tolist'):
            atoms = atoms.tolist()
        entry.n_atoms = len(atoms)

        atoms = list(set(atoms))
        atoms.sort()
        entry.atoms = atoms

        self.chemical = root_section.chemical_name
        if not self.chemical:
            self.chemical = config.services.unavailable_value

        self.quantum_computer_system = root_section.quantum_computer_system
        if root_section.quantum_computing_libraries is not None:
            self.quantum_computing_libraries = root_section.quantum_computing_libraries
        self.computation_datetime = root_section.computation_datetime

        quantities = set()

        quantities.add(root_section.m_def.name)
        for _, property_def, _ in root_section.m_traverse():
            quantities.add(property_def.name)

        self.quantities = list(quantities)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The mirror API of the nomad@FAIRDI APIs. Allows to export upload metadata.
'''

from flask import request, send_file
from flask_restplus import Resource, abort, fields

from nomad import processing as proc
from nomad.datamodel import Dataset
from nomad.doi import DOI
from nomad.files import PublicUploadFiles
from nomad import infrastructure

from .api import api
from .auth import authenticate
from .common import upload_route

ns = api.namespace('mirror', description='Export upload (and all calc) metadata.')


mirror_upload_model = api.model('MirrorUpload', {
    'upload_id': fields.String(description='The id of the exported upload'),
    'upload': fields.String(description='The upload metadata as mongoengine json string'),
    'calcs': fields.List(fields.Raw, description='All upload calculation metadata as mongo SON'),
    'datasets': fields.Raw(description='All upload datasets as dict id->mongo SON'),
    'dois': fields.Raw(description='All upload dois as dict id->mongo SON'),
    'upload_files_path': fields.String(description='The path to the local uploads file folder')
})

mirror_query_model = api.model('MirrorQuery', {
    'query': fields.Raw(
        description='Mongoengine query that is used to search for uploads to mirror.')
})

_Dataset = Dataset.m_def.a_mongo.mongo_cls


@ns.route('/')
class MirrorUploadsResource(Resource):
    @api.doc('get_uploads_mirror')
    @api.response(400, 'Bad query')
    @api.marshal_with(
        mirror_upload_model, skip_none=True, code=200, as_list=True,
        description='Uploads exported')
    @api.expect(mirror_query_model)
    @authenticate(admin_only=True)
    def post(self):
        json_data = request.get_json()
        if json_data is None:
            query = {}
        else:
            query = json_data.get('query', {})

        try:
            return [
                {
                    'upload_id': upload.upload_id,
                    'upload_files_path': upload.upload_files.os_path
                }
                for upload in proc.Upload.objects(**query)], 200
        except Exception as e:
            abort(400, message='Could not query mongodb: %s' % str(e))


def _upload_data(upload_id, upload_json, calcs_col, datasets_col, dois_col):
    calcs = []
    datasets = {}
    dois = {}
    for calc in calcs_col.find(dict(upload_id=upload_id)):
        calcs.append(calc)
        for dataset in calc['metadata'].get('datasets', []):
            if dataset not in datasets:
                datasets[dataset] = datasets_col.find_one(dict(_id=dataset))
                doi = datasets[dataset].get('doi', None)
                if doi is not None:
                    doi_obj = dois_col.find_one(dict(_id=doi))
                    if doi_obj is not None:
                        dois[doi] = doi_obj

    return {
        'upload_id': upload_id,
        'upload': upload_json,
        'calcs': calcs,
        'datasets': datasets,
        'dois': dois
    }


@upload_route(ns)
class MirrorUploadResource(Resource):
    @api.response(400, 'Not available for the given upload, e.g. upload not published.')
    @api.response(404, 'The upload does not exist')
    @api.marshal_with(mirror_upload_model, skip_none=True, code=200, description='Upload exported')
    @api.doc('get_upload_mirror')
    @authenticate(admin_only=True)
    def get(self, upload_id):
        '''
        Export upload (and all calc) metadata for mirrors.
        '''
        try:
            upload = proc.Upload.get(upload_id)
        except KeyError:
            abort(404, message='Upload with id %s does not exist.' % upload_id)

        if upload.tasks_running or upload.process_running:
            abort(400, message='Only non processing uploads can be exported')

        upload_data = _upload_data(
            upload.upload_id,
            upload.to_json(),
            calcs_col=proc.Calc._get_collection(),
            datasets_col=_Dataset._get_collection(),
            dois_col=DOI._get_collection())
        upload_data.update(upload_files_path=upload.upload_files.os_path)
        return upload_data, 200


_mirror_files_parser = api.parser()
_mirror_files_parser.add_argument(
    'prefix', type='str', help='File to download archive or raw', location='args')


@upload_route(ns, '/files')
class MirrorFilesResource(Resource):
    @api.doc('download_files_mirror')
    @api.expect(_mirror_files_parser, validate=True)
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.response(404, 'The upload or calculation does not exist')
    @authenticate(admin_only=True)
    def get(self, upload_id):
        '''
        Download archive and raw files for mirrors
        '''
        try:
            args = request.args
            prefix = args.get('prefix')
            assert prefix in ('archive', 'raw')

        except Exception:
            abort(400, message='bad parameter types')

        try:
            upload_files = PublicUploadFiles(upload_id)

            if prefix == 'raw':
                fileobj = upload_files._raw_file_object('public')

            elif prefix == 'archive':
                fileobj = upload_files._msg_file_object('public')

            else:
                abort(400, message='Unsupported prefix.')

            if not fileobj.exists():
                raise KeyError

            return send_file(
                open(fileobj.os_path, 'rb'),
                mimetype='application/binary',
                as_attachment=True,
                cache_timeout=0,
                attachment_filename=fileobj.os_path)

        except KeyError:
            abort(404, message='Upload %s does not exist' % upload_id)


@ns.route('/users')
class MirrorUsersResource(Resource):
    @api.doc('downlod_users_mirror')
    @api.response(400, 'Unsuccessful userlist query')
    @authenticate(admin_only=True)
    def get(self):
        '''
        Download user list for mirrors
        '''
        try:
            users = infrastructure.keycloak.search_user(query='')
            result = dict()
            for user in users:
                credentials = user.m_to_dict()
                credentials.pop('email', None)
                result[user.username] = credentials

            return result, 200

        except Exception:
            abort(400, message='Failed to fetch users')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import re
import io
from os import path

from nomad.datamodel import EntryArchive
from nomad.parsing import FairdiParser
from nomad.units import ureg as units
from nomad.datamodel.metainfo.public import section_run as Run
from nomad.datamodel.metainfo.public import section_scf_iteration as SCF
from nomad.datamodel.metainfo.public import section_system as System
from nomad.datamodel.metainfo.public import section_single_configuration_calculation as SCC
from nomad.datamodel.metainfo.public import section_method as Method
from nomad.datamodel.metainfo.public import section_sampling_method
from nomad.datamodel.metainfo.public import section_XC_functionals as xc_functionals
from nomad.datamodel.metainfo.public import section_eigenvalues
from nomad.parsing.file_parser import UnstructuredTextFileParser, Quantity

from .metainfo.openmx import OpenmxSCC  # pylint: disable=unused-import

'''
This is parser for OpenMX DFT code.
'''

A = (1 * units.angstrom).to_base_units().magnitude

scf_step_parser = UnstructuredTextFileParser(quantities=[
    Quantity('NormRD', r'NormRD=\s*([\d\.]+)', repeats=False),
    Quantity('Uele', r'Uele=\s*([-\d\.]+)', repeats=False)
])

md_step_parser = UnstructuredTextFileParser(quantities=[
    Quantity('SCF', r'   (SCF=.+?Uele=\s*[-\d\.]+)', sub_parser=scf_step_parser, repeats=True),
    Quantity('Utot', r'Utot\.\s+(-?\d+\.\d+)', repeats=False)
])

species_and_coordinates_parser = UnstructuredTextFileParser(quantities=[
    Quantity('atom', r'\s*\d+\s*([A-Za-z]{1,2})\s*([-\d\.]+)\s+([-\d\.]+)\s+([-\d\.]+)\s+[\d\.]+\s*[\d\.]+\s*',
             repeats=True)
])


def convert_eigenvalues(string):
    values = np.loadtxt(io.StringIO(string), dtype=np.float64, usecols=(1, 2)).transpose()
    return values


eigenvalues_parser = UnstructuredTextFileParser(quantities=[
    Quantity('kpoints', r'k1=\s*([-\.\d]+)\s+k2=\s*([-\.\d]+)\s+k3=\s*([-\.\d]+)',
             repeats=True),
    Quantity('eigenvalues', r'((?:\d+\s+[-\.\d]+\s+[-\.\d]+\s+)+)',
             str_operation=convert_eigenvalues,
             repeats=True)
])

mainfile_parser = UnstructuredTextFileParser(quantities=[
    Quantity('program_version', r'This calculation was performed by OpenMX Ver. ([\d\.]+)\s*', repeats=False),
    Quantity(
        'md_step', r'(SCF history at MD[\s\S]+?Chemical potential \(Hartree\)\s+[-\d\.]+)',
        sub_parser=md_step_parser,
        repeats=True),
    Quantity(
        'atoms', r'<Atoms.SpeciesAndCoordinates([\s\S]+)Atoms.SpeciesAndCoordinates>',
        sub_parser=species_and_coordinates_parser,
        repeats=False),
    Quantity(
        'input_lattice_vectors', r'(?i)<Atoms.UnitVectors\s+((?:-?\d+\.\d+\s+)+)Atoms.UnitVectors>',
        repeats=False),
    Quantity('scf.XcType', r'scf.XcType\s+(\S+)', repeats=False),
    Quantity('scf.SpinPolarization', r'scf.SpinPolarization\s+(\S+)', repeats=False),
    Quantity('Atoms.SpeciesAndCoordinates.Unit',
             r'(?i)Atoms.SpeciesAndCoordinates.Unit\s+([a-z]{2,4})', repeats=False),
    Quantity('Atoms.UnitVectors.Unit',
             r'(?i)Atoms.UnitVectors.Unit\s+([a-z]{2,3})', repeats=False),
    Quantity('scf.Hubbard.U', r'(?i)scf.Hubbard.U\s+(on|off)', repeats=False),
    Quantity('MD.Type', r'(?i)MD\.Type\s+([a-z_\d]{3,6})', repeats=False),
    Quantity('MD.Opt.criterion', r'(?i)MD\.Opt\.criterion\s+([\d\.e-]+)', repeats=False),
    Quantity('scf.maxIter', r'scf.maxIter\s+(\d+)', repeats=False),
    Quantity('scf.criterion', r'scf.criterion\s+([-\.eE\d]+)', repeats=False),
    Quantity('scf.ElectronicTemperature', r'scf.ElectronicTemperature\s+(\S+)', repeats=False),
    Quantity('have_timing', r'Computational Time \(second\)([\s\S]+)Others.+', repeats=False),
    Quantity(
        'eigenvalues', r'Eigenvalues \(Hartree\) \S+ SCF KS-eq\.\s+\*{59}\s*\*{59}([^*]+)',
        sub_parser=eigenvalues_parser,
        repeats=False)
])


def read_md_file(md_file):
    result = []
    with open(md_file, 'r') as f:
        cell_vectors_re = re.compile(r'Cell_Vectors=((?:\s+-?\d+\.\d+)+)')
        temperature_re = re.compile(r'Temperature=\s+(\d+\.\d+)')
        for line in f:
            line_list = line.split()
            if len(line_list) == 1:
                step_header = True
                natoms = int(line_list[0])
                result.append({'species': [],
                               'positions': np.empty((natoms, 3), dtype=np.float64),
                               'velocities': np.empty((natoms, 3), dtype=np.float64),
                               'forces': np.empty((natoms, 3), dtype=np.float64)})
                atomindex = 0
            elif step_header:
                cell_vectors = cell_vectors_re.search(line)
                if cell_vectors is not None:
                    cell_vectors = [float(v) for v in cell_vectors.group(1).split()]
                    result[-1]['cell_vectors'] = np.array(cell_vectors).reshape(3, 3)
                temperature = temperature_re.search(line)
                if temperature is not None:
                    temperature = temperature.group(1)
                    result[-1]['temperature'] = temperature
                step_header = False
            else:
                result[-1]['positions'][atomindex][0:3] = [
                    float(val) for val in line_list[1:4]]
                result[-1]['forces'][atomindex][0:3] = [
                    float(val) for val in line_list[4:7]]
                result[-1]['velocities'][atomindex][0:3] = [
                    float(val) for val in line_list[7:10]]
                result[-1]['species'].append(line_list[0])
                atomindex += 1
    f.close()
    return result


def parse_md_file(i, mdfile_md_steps, system):
    cell = mdfile_md_steps[i].get('cell_vectors')
    system.lattice_vectors = cell * units.angstrom
    system.configuration_periodic_dimensions = [True, True, True]
    positions = mdfile_md_steps[i].get('positions')
    system.atom_positions = positions * units.angstrom
    system.atom_labels = mdfile_md_steps[i].get('species')
    system.velocities = mdfile_md_steps[i].get('velocities') * units.meter / units.second


def parse_structure(system, logger):
    atoms_units = mainfile_parser.get('Atoms.SpeciesAndCoordinates.Unit')
    lattice_vectors = mainfile_parser.get('input_lattice_vectors')
    lattice_units = mainfile_parser.get('Atoms.UnitVectors.Unit')
    atoms = mainfile_parser.get('atoms').get('atom')

    if atoms is not None and lattice_vectors is not None:
        lattice_vectors = np.array(lattice_vectors).reshape(3, 3)
        if atoms_units is not None:
            if lattice_units.lower() == 'au':
                lattice_vectors = lattice_vectors * units.bohr
                lattice_units = units.bohr
            elif lattice_units.lower() == 'ang':
                lattice_vectors = lattice_vectors * units.angstrom
                lattice_units = units.angstrom
        else:
            lattice_vectors = lattice_vectors * units.angstrom
        atom_positions = [a[1:4] for a in atoms]
        if atoms_units is not None:
            if atoms_units.lower() == 'au':
                atom_positions = atom_positions * units.bohr
            if atoms_units.lower() == 'ang':
                atom_positions = atom_positions * units.angstrom
            elif atoms_units.lower() == 'frac':
                # There are some problems with pint here so simple matrix vector multiplications
                # doesn't work.
                atom_positions = np.array([np.array(pos).dot(lattice_vectors)
                                          for pos in atom_positions]) * lattice_units
        else:
            # default unit is angstrom
            atom_positions = atom_positions * units.angstrom
        system.atom_positions = atom_positions
        system.lattice_vectors = lattice_vectors
        system.configuration_periodic_dimensions = [True, True, True]
        system.atom_labels = [a[0] for a in atoms]
    else:
        logger.warning('Failed to parse the input structure.')


class OpenmxParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/openmx', code_name='OpenMX', code_homepage='http://www.openmx-square.org/',
            mainfile_mime_re=r'(text/.*)',
            mainfile_name_re=r'.*\.out$',
            mainfile_contents_re=(r'^\*{59}\s+\*{59}\s+This calculation was performed by OpenMX'),
        )

    def parse_sampling_method(self):
        md_type = mainfile_parser.get('MD.Type')
        md_types_list = [
            # FIXME: handle the various OptCx methods with constraints
            ['OPT', 'geometry_optimization', 'steepest_descent'],
            ['DIIS', 'geometry_optimization', 'diis'],
            ['BFGS', 'geometry_optimization', 'bfgs'],
            # FIXME: not in https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/-/wikis/metainfo/geometry-optimization-method
            ['RF', 'geometry_optimization', 'rf'],
            # FIXME: not in https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/-/wikis/metainfo/geometry-optimization-method
            ['EF', 'geometry_optimization', 'ef'],
            ['NVE', 'molecular_dynamics', 'NVE'],
            ['NVT_VS', 'molecular_dynamics', 'NVT'],
            ['NVT_NH', 'molecular_dynamics', 'NVT'],
        ]
        if md_type is not None and 'nomd' not in md_type.lower():
            md_type = md_type.upper()
            sampling_method = self.archive.section_run[-1].m_create(section_sampling_method)
            for current_md_type in md_types_list:
                if current_md_type[0] in md_type:
                    sampling_method.sampling_method = current_md_type[1]
                    if current_md_type[1] == 'geometry_optimization':
                        sampling_method.geometry_optimization_method = current_md_type[2]
                        criterion = mainfile_parser.get('MD.Opt.criterion')
                        if criterion is not None:
                            sampling_method.geometry_optimization_threshold_force = (
                                criterion * units.hartree / units.bohr)
                        else:
                            sampling_method.geometry_optimization_threshold_force = (
                                1e-4 * units.hartree / units.bohr)
                    else:
                        sampling_method.ensemble_type = current_md_type[2]

    def parse_method(self):
        method = self.archive.section_run[-1].m_create(Method)
        method.electronic_structure_method = 'DFT'
        # FIXME: add some testcase for DFT+U
        scf_hubbard_u = mainfile_parser.get('scf.Hubbard.U')
        if scf_hubbard_u is not None:
            if scf_hubbard_u.lower() == 'on':
                method.electronic_structure_method = 'DFT+U'

        xc_functional_dictionary = {
            'GGA-PBE': ['GGA_C_PBE', 'GGA_X_PBE'],
            'LDA': ['LDA_X', 'LDA_C_PZ'],
            'LSDA-CA': ['LDA_X', 'LDA_C_PZ'],
            'LSDA-PW': ['LDA_X', 'LDA_C_PW'],
            None: ['LDA_X', 'LDA_C_PZ']
        }
        scf_xctype = mainfile_parser.get('scf.XcType')
        for xc in xc_functional_dictionary[scf_xctype]:
            method.m_create(xc_functionals).XC_functional_name = xc

        spinpol = mainfile_parser.get('scf.SpinPolarization')
        if spinpol.lower() == 'on':
            method.number_of_spin_channels = 2
            self.spinpolarized = True
        else:
            method.number_of_spin_channels = 1
            self.spinpolarized = False

        method.smearing_kind = 'fermi'
        scf_ElectronicTemperature = mainfile_parser.get('scf.ElectronicTemperature')
        if scf_ElectronicTemperature is not None:
            method.smearing_width = (scf_ElectronicTemperature * units.kelvin * units.k).to_base_units().magnitude
        else:
            method.smearing_width = (300 * units.kelvin * units.k).to_base_units().magnitude

        scf_maxiter = mainfile_parser.get('scf.maxIter')
        if scf_maxiter is not None:
            method.scf_max_iteration = scf_maxiter

        scf_criterion = mainfile_parser.get('scf.criterion')
        if scf_criterion is not None:
            method.scf_threshold_energy_change = scf_criterion * units.hartree

    def parse_eigenvalues(self):
        eigenvalues = self.archive.section_run[-1].section_single_configuration_calculation[-1].m_create(section_eigenvalues)
        eigenvalues.eigenvalues_kind = 'normal'
        values = mainfile_parser.get('eigenvalues')
        if values is not None:
            kpoints = values.get('kpoints')
            if kpoints is not None:
                eigenvalues.eigenvalues_kpoints = kpoints
                eigenvalues.number_of_eigenvalues_kpoints = len(kpoints)
            else:
                eigenvalues.eigenvalues_kpoints = [[0, 0, 0]]
                eigenvalues.number_of_eigenvalues_kpoints = 1
            values = values.get('eigenvalues')
            if values is not None:
                if self.spinpolarized:
                    eigenvalues.eigenvalues_values = np.stack(values, axis=1) * units.hartree
                else:
                    values = [i[0:1] for i in values]
                    eigenvalues.eigenvalues_values = np.stack(values, axis=1) * units.hartree
                eigenvalues.number_of_eigenvalues = len(values[0][0])

    def parse(self, mainfile: str, archive: EntryArchive, logger):
        self.archive = archive

        # Use the previously defined parsers on the given mainfile
        mainfile_parser.mainfile = mainfile
        mainfile_parser.parse()
        del mainfile_parser._file_handler

        # Get system from the MD file
        md_file = path.splitext(mainfile)[0] + '.md'
        if path.isfile(md_file):
            mdfile_md_steps = read_md_file(md_file)
        else:
            mdfile_md_steps = None

        # Some basic values
        run = archive.m_create(Run)
        run.program_name = 'OpenMX'
        run.program_version = str(mainfile_parser.get('program_version'))
        run.program_basis_set_type = "Numeric AOs"

        have_timing = mainfile_parser.get('have_timing')
        if have_timing is not None:
            run.run_clean_end = True
        else:
            run.run_clean_end = False

        self.parse_sampling_method()

        self.parse_method()

        mainfile_md_steps = mainfile_parser.get('md_step')
        if mainfile_md_steps is not None:
            n_md_steps = len(mainfile_md_steps)
        n_mdfile_md_steps = 0
        if mdfile_md_steps is not None:
            n_mdfile_md_steps = len(mdfile_md_steps)
        # Do some consistency checks between the out and md file.
        ignore_md_file = False
        if n_mdfile_md_steps > n_md_steps:
            # This can happen when user runs two calculations in the same directory.
            # In that case the out file contains the latter calculation but the md file
            # would contain both calculations, so just take the corresponding number
            # of steps from the end of the file.
            mdfile_md_steps = mdfile_md_steps[-n_md_steps:]
        elif n_mdfile_md_steps < n_md_steps:
            # This is unlikely, but signals a problem with the md file, so just
            # ignore it.
            ignore_md_file = True
            logger.warning(".md file does not contain enough MD steps")

        if mainfile_md_steps is not None:
            for i, md_step in enumerate(mainfile_md_steps):
                system = run.m_create(System)
                if not ignore_md_file:
                    parse_md_file(i, mdfile_md_steps, system)
                elif i == 0:
                    # Get the initial position from out file as a fallback if the md file is missing.
                    parse_structure(system, logger)

                scc = run.m_create(SCC)
                scc.single_configuration_calculation_to_system_ref = system
                scc.single_configuration_to_calculation_method_ref = run.section_method[-1]
                scf_steps = md_step.get('SCF')
                if scf_steps is not None:
                    for scf_step in scf_steps:
                        scf = scc.m_create(SCF)
                        u_ele = scf_step.get('Uele')
                        if u_ele is not None:
                            scf.energy_sum_eigenvalues_scf_iteration = u_ele * units.hartree
                u_tot = md_step.get('Utot')
                if u_tot is not None:
                    scc.energy_total = u_tot * units.hartree
                if not ignore_md_file:
                    forces = mdfile_md_steps[i].get('forces')
                    if forces is not None:
                        scc.atom_forces = forces * units.hartree / units.bohr
                    temperature = mdfile_md_steps[i].get('temperature')
                    if temperature is not None:
                        scc.temperature = temperature * units.kelvin

        self.parse_eigenvalues()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This module represents calculations in elastic search.
'''

from typing import cast, Iterable, Dict, List, Any
from elasticsearch_dsl import Search, Q, A, analyzer, tokenizer
import elasticsearch.helpers
from elasticsearch.exceptions import NotFoundError, RequestError, TransportError
from datetime import datetime
import json

from nomad.datamodel.material import Material
from nomad import config, datamodel, infrastructure, utils
from nomad.metainfo.search_extension import (  # pylint: disable=unused-import
    search_quantities, metrics, order_default_quantities, groups)
from nomad.app.v1 import models as api_models
from nomad.app.v1.models import (
    EntryPagination, PaginationResponse, Query, MetadataRequired, SearchResponse, Aggregation,
    Statistic, StatisticResponse, AggregationOrderType, AggregationResponse, AggregationDataItem)


_entry_metadata_defaults = {
    quantity.name: quantity.default
    for quantity in datamodel.EntryMetadata.m_def.quantities  # pylint: disable=not-an-iterable
    if quantity.default not in [None, [], False, 0]
}

_all_author_quantities = [
    quantity.name
    for quantity in datamodel.EntryMetadata.m_def.all_quantities.values()
    if quantity.type in [datamodel.user_reference, datamodel.author_reference]]


def _es_to_entry_dict(hit, required: MetadataRequired) -> Dict[str, Any]:
    '''
    Elasticsearch entry metadata does not contain default values, if a metadata is not
    set. This will add default values to entry metadata in dict form obtained from
    elasticsearch.
    '''
    entry_dict = hit.to_dict()
    for key, value in _entry_metadata_defaults.items():
        if key not in entry_dict:
            if required is not None:
                if required.exclude and key in required.exclude:
                    continue
                if required.include and key not in required.include:
                    continue

            entry_dict[key] = value

    for author_quantity in _all_author_quantities:
        authors = entry_dict.get(author_quantity)
        if authors is None:
            continue
        if isinstance(authors, dict):
            authors = [authors]
        for author in authors:
            if 'email' in author:
                del(author['email'])

    return entry_dict


path_analyzer = analyzer(
    'path_analyzer',
    tokenizer=tokenizer('path_tokenizer', 'pattern', pattern='/'))


class AlreadyExists(Exception): pass


class ElasticSearchError(Exception): pass


class AuthenticationRequiredError(Exception): pass


class ScrollIdNotFound(Exception): pass


class InvalidQuery(Exception): pass


entry_document = datamodel.EntryMetadata.m_def.a_elastic.document
material_document = Material.m_def.a_elastic.document

for domain in datamodel.domains:
    order_default_quantities.setdefault(domain, order_default_quantities.get('__all__'))


def delete_upload(upload_id):
    ''' Delete all entries with given ``upload_id`` from the index. '''
    index = entry_document._default_index()
    Search(index=index).query('match', upload_id=upload_id).delete()


def delete_entry(calc_id):
    ''' Delete the entry with the given ``calc_id`` from the index. '''
    index = entry_document._default_index()
    Search(index=index).query('match', calc_id=calc_id).delete()


def publish(calcs: Iterable[datamodel.EntryMetadata]) -> None:
    ''' Update all given calcs with their metadata and set ``publish = True``. '''
    def elastic_updates():
        for calc in calcs:
            entry = calc.a_elastic.create_index_entry()
            entry.published = True
            entry = entry.to_dict(include_meta=True)
            source = entry.pop('_source')
            entry['doc'] = source
            entry['_op_type'] = 'update'
            yield entry

    elasticsearch.helpers.bulk(infrastructure.elastic_client, elastic_updates())
    refresh()


def index_all(calcs: Iterable[datamodel.EntryMetadata], do_refresh=True) -> None:
    '''
    Adds all given calcs with their metadata to the index.

    Returns:
        Number of failed entries.
    '''
    def elastic_updates():
        for calc in calcs:
            try:
                entry = calc.a_elastic.create_index_entry()
                entry = entry.to_dict(include_meta=True)
                entry['_op_type'] = 'index'

                yield entry

            except Exception as e:
                utils.get_logger(__name__).error(
                    'could not create index doc', exc_info=e,
                    upload_id=calc.upload_id, calc_id=calc.calc_id)

    _, failed = elasticsearch.helpers.bulk(infrastructure.elastic_client, elastic_updates(), stats_only=True)

    if do_refresh:
        refresh()

    return failed


def refresh():
    infrastructure.elastic_client.indices.refresh(config.elastic.index_name)


def _owner_es_query(owner: str, user_id: str = None):
    if owner == 'all':
        q = Q('term', published=True)
        if user_id is not None:
            q = q | Q('term', owners__user_id=user_id)
    elif owner == 'public':
        q = Q('term', published=True) & Q('term', with_embargo=False)
    elif owner == 'visible':
        q = Q('term', published=True) & Q('term', with_embargo=False)
        if user_id is not None:
            q = q | Q('term', owners__user_id=user_id)
    elif owner == 'shared':
        if user_id is None:
            raise AuthenticationRequiredError('Authentication required for owner value shared.')

        q = Q('term', owners__user_id=user_id)
    elif owner == 'user':
        if user_id is None:
            raise AuthenticationRequiredError('Authentication required for owner value user.')

        q = Q('term', uploader__user_id=user_id)
    elif owner == 'staging':
        if user_id is None:
            raise AuthenticationRequiredError('Authentication required for owner value user')
        q = Q('term', published=False) & Q('term', owners__user_id=user_id)
    elif owner == 'admin':
        if user_id is None or not datamodel.User.get(user_id=user_id).is_admin:
            raise AuthenticationRequiredError('This can only be used by the admin user.')
        q = None
    else:
        raise KeyError('Unsupported owner value')

    if q is not None:
        return q
    return Q()


class SearchRequest:
    '''
    Represents a search request and allows to execute that request.
    It allows to compose the following features: a query;
    statistics (metrics and aggregations); quantity values; scrolling, pagination for entries;
    scrolling for quantity values.

    The query part filters NOMAD data before the other features come into effect. There
    are specialized methods for configuring the :func:`owner` and :func:`time_range` queries.
    Quantity's can be search for by setting them as attributes.

    The aggregations for statistics can be requested for pre-configured quantities. These
    bucket aggregations come with a metric calculated for each each possible
    quantity value.

    The other possible form of aggregations, allows to get quantity values as results
    (e.g. get all datasets, get all users, etc.). Each value can be accompanied by metrics
    (over all entries with that value) and an example value.

    Of course, searches can return a set of search results. Search objects can be
    configured with pagination or scrolling for these results. Pagination is the default
    and also allows ordering of results. Scrolling can be used if all entries need to be
    'scrolled through'. This might be necessary, since elastic search has limits on
    possible pages (e.g. 'from' must by smaller than 10000). On the downside, there is no
    ordering on scrolling.

    There is also scrolling for quantities to go through all quantity values. There is no
    paging for aggregations.
    '''
    def __init__(self, domain: str = config.meta.default_domain, query=None):
        self._domain = domain
        self._query = query
        self._search = Search(index=config.elastic.index_name)
        self._required = None

    def domain(self, domain: str = None):
        '''
        Applies the domain of this request to the query. Allows to optionally update
        the domain of this request.
        '''
        if domain is not None:
            self._domain = domain

        self.q = self.q & Q('term', domain=self._domain)
        return self

    def owner(self, owner_type: str = 'all', user_id: str = None):
        '''
        Uses the query part of the search to restrict the results based on the owner.
        The possible types are: ``all`` for all calculations; ``public`` for
        calculations visible by everyone, excluding embargo-ed entries and entries only visible
        to the given user; ``visible`` all data that is visible by the user, excluding
        embargo-ed entries from other users; ``user`` for all calculations of to the given
        user; ``staging`` for all calculations in staging of the given user.

        Arguments:
            owner_type: The type of the owner query, see above.
            user_id: The 'owner' given as the user's unique id.

        Raises:
            KeyError: If the given owner_type is not supported
            ValueError: If the owner_type requires a user but none is given, or the
                given user is not allowed to use the given owner_type.
        '''
        self.q &= _owner_es_query(owner=owner_type, user_id=user_id)
        return self

    def search_parameters(self, **kwargs):
        '''
        Configures the existing query with additional search parameters. Kwargs are
        interpreted as key value pairs. Keys have to coresspond to valid entry quantities
        in the domain's (DFT calculations) datamodel. Alternatively search parameters
        can be set via attributes.
        '''
        for name, value in kwargs.items():
            self.search_parameter(name, value)

        return self

    def _search_parameter_to_es(self, name, value):
        quantity = search_quantities[name]

        if quantity.many and not isinstance(value, list):
            value = [value]

        if quantity.many_or and isinstance(value, List):
            return Q('terms', **{quantity.search_field: value})

        if quantity.derived:
            if quantity.many and not isinstance(value, list):
                value = [value]
            value = quantity.derived(value)

        if isinstance(value, list):
            values = value
        else:
            values = [value]

        return Q('bool', must=[
            Q('match', **{quantity.search_field: item})
            for item in values])

    def search_parameter(self, name, value):
        self.q &= self._search_parameter_to_es(name, value)
        return self

    def query(self, query):
        ''' Adds the given query as a 'and' (i.e. 'must') clause to the request. '''
        self._query &= query

        return self

    def query_expression(self, expression) -> 'SearchRequest':

        bool_operators = {'$and': 'must', '$or': 'should', '$not': 'must_not'}
        comp_operators = {'$%s' % op: op for op in ['gt', 'gte', 'lt', 'lte']}

        def _to_es(key, values):
            if key in bool_operators:
                if isinstance(values, dict):
                    values = [values]
                assert isinstance(values, list), 'bool operator requires a list of dicts or dict'
                child_es_queries = [
                    _to_es(child_key, child_value)
                    for child_query in values
                    for child_key, child_value in child_query.items()]
                return Q('bool', **{bool_operators[key]: child_es_queries})

            if key in comp_operators:
                assert isinstance(values, dict), 'comparison operator requires a dict'
                assert len(values) == 1, 'comparison operator requires exactly one quantity'
                quantity_name, value = next(iter(values.items()))
                quantity = search_quantities.get(quantity_name)
                assert quantity is not None, 'quantity %s does not exist' % quantity_name
                return Q('range', **{quantity.search_field: {comp_operators[key]: value}})

            try:
                return self._search_parameter_to_es(key, values)
            except KeyError:
                assert False, 'quantity %s does not exist' % key

        if len(expression) == 0:
            self.q &= Q()
        else:
            self.q &= Q('bool', must=[_to_es(key, value) for key, value in expression.items()])

        return self

    def time_range(self, start: datetime, end: datetime):
        ''' Adds a time range to the query. '''
        if start is None and end is None:
            return self

        if start is None:
            start = datetime.fromtimestamp(0)
        if end is None:
            end = datetime.utcnow()

        self.q &= Q('range', upload_time=dict(gte=start, lte=end))

        return self

    @property
    def q(self):
        ''' The underlying elasticsearch_dsl query object '''
        if self._query is None:
            return Q('match_all')
        else:
            return self._query

    @q.setter
    def q(self, q):
        self._query = q

    def totals(self, metrics_to_use: List[str] = []):
        '''
        Configure the request to return overall totals for the given metrics.

        The statics are returned with the other quantity statistics under the pseudo
        quantity name 'total'. 'total' contains the pseudo value 'all'. It is used to
        store the metrics aggregated over all entries in the search results.
        '''
        self._add_metrics(self._search.aggs, metrics_to_use)
        return self

    def statistics(self, statistics: List[str], metrics_to_use: List[str] = []):
        '''
        Configures the domain's default statistics.
        '''
        for statistic in statistics:
            search_quantity = search_quantities[statistic]
            statistic_order = search_quantity.statistic_order
            self.statistic(
                search_quantity.qualified_name,
                search_quantity.statistic_size,
                metrics_to_use=metrics_to_use,
                order={statistic_order: 'asc' if statistic_order == '_key' else 'desc'})

        return self

    def statistic(
            self, quantity_name: str, size: int, metrics_to_use: List[str] = [],
            order: Dict[str, str] = dict(_key='asc'), include: str = None):
        '''
        This can be used to display statistics over the searched entries and allows to
        implement faceted search on the top values for each quantity.

        The metrics contain overall and per quantity value sums of code runs (calcs),
        unique code runs, datasets, and additional domain specific metrics
        (e.g. total energies, and unique geometries for DFTcalculations). The quantities
        that can be aggregated to metrics are defined in module:`datamodel`. Aggregations
        and respective metrics are calculated for aggregations given in ``aggregations``
        and metrics in ``aggregation_metrics``. As a pseudo aggregation ``total_metrics``
        are calculation over all search results. The ``aggregations`` gives tuples of
        quantities and default aggregation sizes.

        The search results will contain a dictionary ``statistics``. This has a key
        for each configured quantity. Each quantity key will hold a dict
        with a key for each quantity value. Each quantity value key will hold a dict
        with a key for each metric. The values will be the actual aggregated metric values.

        Arguments:
            quantity_name: The quantity to aggregate statistics for. Only works on *keyword* field.
            metrics_to_use: The metrics calculated over the aggregations. Can be
                ``unique_code_runs``, ``datasets``, other domain specific metrics.
                The basic doc_count metric ``code_runs`` is always given.
            order: The order dictionary is passed to the elastic search aggregation.
            include:
                Uses an regular expression in ES to only return values that include
                the given substring.
        '''
        quantity = search_quantities[quantity_name]
        terms_kwargs = {}
        if include is not None:
            terms_kwargs['include'] = '.*%s.*' % include
        terms = A('terms', field=quantity.search_field, size=size, order=order, **terms_kwargs)

        buckets = self._search.aggs.bucket('statistics:%s' % quantity_name, terms)
        self._add_metrics(buckets, metrics_to_use)

        return self

    def _add_metrics(self, parent=None, metrics_to_use: List[str] = []):
        if parent is None:
            parent = self._search.aggs

        for metric in metrics_to_use:
            metric_quantity = metrics[metric]
            field = metric_quantity.search_field
            parent.metric(
                'metric:%s' % metric_quantity.metric_name,
                A(metric_quantity.metric, field=field))

    def date_histogram(self, metrics_to_use: List[str] = [], interval: str = '1M'):
        '''
        Adds a date histogram on the given metrics to the statistics part.
        '''
        histogram = A('date_histogram', field='upload_time', interval=interval, format='yyyy-MM-dd')
        self._add_metrics(self._search.aggs.bucket('statistics:date_histogram', histogram), metrics_to_use)

        return self

    def quantities(self, **kwargs):
        '''
        Shorthand for adding multiple quantities. See :func:`quantity`. Keywork argument
        keys are quantity name, values are tuples of size and after value.
        '''
        for name, spec in kwargs:
            size, after = spec
            self.quantity(name, after=after, size=size)

        return self

    def quantity(
            self, name, size=100, after=None, examples=0, examples_source=None,
            order_by: str = None, order: str = 'desc'):
        '''
        Adds a requests for values of the given quantity.
        It allows to scroll through all values via elasticsearch's
        composite aggregations. The response will contain the quantity values and
        an example entry for each value.

        This can be used to implement continues scrolling through authors, datasets,
        or uploads within the searched entries.

        If one or more quantities are specified,
        the search results will contain a dictionary ``quantities``. The keys are quantity
        name the values dictionary with 'after' and 'values' key.
        The 'values' key holds a dict with all the values as keys and their entry count
        as values (i.e. number of entries with that value).

        Arguments:
            name: The quantity name. Must be in :data:`quantities`.
            after: The 'after' value allows to scroll over various requests, by providing
                the 'after' value of the last search. The 'after' value is part of the
                response. Use ``None`` in the first request.
            size:
                The size gives the ammount of maximum values in the next scroll window.
                If the size is None, a maximum of 100 quantity values will be requested.
            examples:
                Number of results to return that has each value
            order_by:
                A sortable quantity that should be used to order. By default, the max of each
                value bucket is used.
            order:
                "desc" or "asc"
        '''
        if size is None:
            size = 100

        quantity = search_quantities[name]
        terms = A('terms', field=quantity.search_field)

        # We are using elastic searchs 'composite aggregations' here. We do not really
        # compose aggregations, but only those pseudo composites allow us to use the
        # 'after' feature that allows to scan through all aggregation values.
        if order_by is None:
            composite = dict(sources={name: terms}, size=size)
        else:
            sort_terms = A('terms', field=order_by, order=order)
            composite = dict(sources=[{order_by: sort_terms}, {name: terms}], size=size)
        if after is not None:
            if order_by is None:
                composite['after'] = {name: after}
            else:
                composite['after'] = {order_by: after, name: ''}

        composite_agg = self._search.aggs.bucket('quantity:%s' % name, 'composite', **composite)

        if examples > 0:
            kwargs: Dict[str, Any] = {}
            if examples_source is not None:
                kwargs.update(_source=dict(includes=examples_source))

            composite_agg.metric('examples', A('top_hits', size=examples, **kwargs))

        return self

    def global_statistics(self):
        '''
        Adds general statistics to the request. The results will have a key called
        global_statistics.
        '''
        self.owner('public')
        self._search.aggs.metric(
            'global_statistics:n_entries', A('value_count', field='calc_id'))
        self._search.aggs.metric(
            'global_statistics:n_uploads', A('cardinality', field='upload_id'))
        self._search.aggs.metric(
            'global_statistics:n_calculations', A('sum', field='dft.n_calculations'))
        self._search.aggs.metric(
            'global_statistics:n_quantities', A('sum', field='dft.n_quantities'))
        self._search.aggs.metric(
            'global_statistics:n_materials', A('cardinality', field='encyclopedia.material.material_id'))

        return self

    def exclude(self, *args):
        ''' Exclude certain elastic fields from the search results. '''
        self._search = self._search.source(excludes=args)
        self._required = MetadataRequired(exclude=args)
        return self

    def include(self, *args):
        ''' Include only the given fields in the search results. '''
        self._search = self._search.source(includes=args)
        return self

    def execute(self):
        '''
        Executes without returning actual results. Only makes sense if the request
        was configured for statistics or quantity values.
        '''
        search = self._search.query(self.q)[0:0]
        response = search.execute()
        return self._response(response)

    def execute_scan(self, order_by: str = None, order: int = -1, **kwargs):
        '''
        This execute the search as scan. The result will be a generator over the found
        entries. Everything but the query part of this object, will be ignored.
        '''
        search = self._search.query(self.q)

        if order_by is not None:
            order_by_quantity = search_quantities[order_by]

            if order == 1:
                search = search.sort(order_by_quantity.search_field)
            else:
                search = search.sort('-%s' % order_by_quantity.search_field)  # pylint: disable=no-member

            search = search.params(preserve_order=True)

        for hit in search.params(**kwargs).scan():
            yield _es_to_entry_dict(hit, self._required)

    def execute_paginated(
            self,
            page: int = 1, per_page=10, page_offset: int = None,
            order_by: str = None, order: int = -1):
        '''
        Executes the search and returns paginated results. Those are sorted.

        Arguments:
            page: The requested page, starts with 1.
            per_page: The number of entries per page.
            page_offset: Instead of a page number, use this absolute offset.
            order_by: The quantity to order by.
            order: -1 or 1 for descending or ascending order.
        '''
        if order_by is None:
            order_by_quantity = order_default_quantities[self._domain]
        else:
            order_by_quantity = search_quantities[order_by]

        search = self._search.query(self.q)

        if order == 1:
            search = search.sort(order_by_quantity.search_field)
        else:
            search = search.sort('-%s' % order_by_quantity.search_field)  # pylint: disable=no-member

        if page_offset is not None:
            search = search[page_offset: page_offset + per_page]  # pylint: disable=unsubscriptable-object
        else:
            search = search[(page - 1) * per_page: page * per_page]  # pylint: disable=unsubscriptable-object

        es_result = search.execute()

        result = self._response(es_result, with_hits=True)

        if page_offset is not None:
            result.update(pagination=dict(
                total=result['total'],
                page_offset=page_offset, per_page=per_page))
        else:
            result.update(
                pagination=dict(total=result['total'], page=page, per_page=per_page))
        return result

    def execute_scrolled(
            self, scroll_id: str = None, size: int = 1000, scroll: str = u'5m',
            order_by: str = None, order: int = -1):
        '''
        Executes a scrolling search. based on ES scroll API. Pagination is replaced with
        scrolling, no ordering is available, no statistics, no quantities will be provided.

        Scrolling is done by calling this function again and again with the same ``scroll_id``.
        Each time, this function will return the next batch of search results. If the
        ``scroll_id`` is not available anymore, a new ``scroll_id`` is assigned and scrolling
        starts from the beginning again.

        The response will contain a 'scroll' part with attributes 'total', 'scroll_id',
        and 'size'.

        Arguments:
            scroll_id: The scroll id to receive the next batch from. None will create a new
                scroll.
            size: The batch size in number of hits.
            scroll: The time the scroll should be kept alive (i.e. the time between requests
                to this method) in ES time units. Default is 5 minutes.

        TODO support order and order_by
        '''
        es = infrastructure.elastic_client

        if scroll_id is None:
            # initiate scroll
            resp = es.search(  # pylint: disable=E1123
                body=self._search.query(self.q).to_dict(), scroll=scroll, size=size,
                index=config.elastic.index_name)

            scroll_id = resp.get('_scroll_id')
            if scroll_id is None:
                # no results for search query
                return dict(scroll=dict(total=0, size=size), results=[])

        else:
            try:
                resp = es.scroll(scroll_id, scroll=scroll)  # pylint: disable=E1123
            except NotFoundError:
                raise ScrollIdNotFound()

        total = resp['hits']['total']
        results = list(hit['_source'] for hit in resp['hits']['hits'])

        # since we are using the low level api here, we should check errors
        if resp["_shards"]["successful"] < resp["_shards"]["total"]:
            utils.get_logger(__name__).error('es operation was unsuccessful on at least one shard')
            raise ElasticSearchError('es operation was unsuccessful on at least one shard')

        if len(results) == 0:
            es.clear_scroll(body={'scroll_id': [scroll_id]}, ignore=(404, ))  # pylint: disable=E1123
            scroll_id = None

        scroll_info = dict(total=total, size=size, scroll=True)
        if scroll_id is not None:
            scroll_info.update(scroll_id=scroll_id)

        return dict(scroll=scroll_info, results=results)

    def execute_aggregated(
            self, after: str = None, per_page: int = 1000, includes: List[str] = None):
        '''
        Uses a composite aggregation on top of the search to go through the result
        set. This allows to go arbirarely deep without using scroll. But, it will
        only return results with ``upload_id``, ``calc_id`` and the given
        quantities. The results will be 'ordered' by ``upload_id``.

        Arguments:
            after: The key that determines the start of the current page. This after
                key is returned with each response. Use None (default) for the first
                request.
            per_page: The size of each page.
            includes: A list of quantity names that should be returned in addition to
                ``upload_id`` and ``calc_id``.
        '''
        upload_id_agg = A('terms', field="upload_id")
        calc_id_agg = A('terms', field="calc_id")

        composite = dict(
            sources=[dict(upload_id=upload_id_agg), dict(calc_id=calc_id_agg)],
            size=per_page)

        if after is not None:
            upload_id, calc_id = after.split(':')
            composite['after'] = dict(upload_id=upload_id, calc_id=calc_id)

        composite_agg = self._search.aggs.bucket('ids', 'composite', **composite)
        if includes is not None:
            composite_agg.metric('examples', A('top_hits', size=1, _source=dict(includes=includes)))

        search = self._search.query(self.q)[0:0]
        response = search.execute()

        ids = response['aggregations']['ids']
        if 'after_key' in ids:
            after_dict = ids['after_key']
            after = '%s:%s' % (after_dict['upload_id'], after_dict['calc_id'])
        else:
            after = None

        id_agg_info = dict(total=response['hits']['total'], after=after, per_page=per_page)

        def transform_result(es_result):
            result = dict(
                upload_id=es_result['key']['upload_id'],
                calc_id=es_result['key']['calc_id'])

            if includes is not None:
                source = es_result['examples']['hits']['hits'][0]['_source']
                for key in source:
                    result[key] = source[key]

            return result

        results = [
            transform_result(item) for item in ids['buckets']]

        return dict(aggregation=id_agg_info, results=results)

    def _response(self, response, with_hits: bool = False) -> Dict[str, Any]:
        '''
        Prepares a response object covering the total number of results, hits, statistics,
        and quantities. Other aspects like pagination and scrolling have to be added
        elsewhere.
        '''
        result: Dict[str, Any] = dict()
        aggs = response.aggregations.to_dict()

        # total
        total = response.hits.total if hasattr(response, 'hits') else 0
        result.update(total=total)

        # hits
        if len(response.hits) > 0 or with_hits:
            result.update(results=[_es_to_entry_dict(hit, self._required) for hit in response.hits])

        # statistics
        def get_metrics(bucket, code_runs):
            result = {}
            # TODO optimize ... go through the buckets not the metrics
            for metric in metrics:
                agg_name = 'metric:%s' % metric
                if agg_name in bucket:
                    result[metric] = bucket[agg_name]['value']
            result.update(code_runs=code_runs)
            return result

        statistics_results = {
            quantity_name[11:]: {
                str(bucket['key']): get_metrics(bucket, bucket['doc_count'])
                for bucket in quantity['buckets']
            }
            for quantity_name, quantity in aggs.items()
            if quantity_name.startswith('statistics:')
        }

        # global statistics
        global_statistics_results = {
            agg_name[18:]: agg.get('value')
            for agg_name, agg in aggs.items()
            if agg_name.startswith('global_statistics:')
        }
        if len(global_statistics_results) > 0:
            result.update(global_statistics=global_statistics_results)

        # totals
        totals_result = get_metrics(aggs, total)
        statistics_results['total'] = dict(all=totals_result)

        if len(statistics_results) > 0:
            result.update(statistics=statistics_results)

        # quantities
        def create_quantity_result(quantity_name, quantity):
            values = {}
            for bucket in quantity['buckets']:
                value = dict(
                    total=bucket['doc_count'])
                if 'examples' in bucket:
                    examples = [hit['_source'] for hit in bucket['examples']['hits']['hits']]
                    value.update(examples=examples)

                values[bucket['key'][quantity_name]] = value

            result = dict(values=values)
            if 'after_key' in quantity:
                after = quantity['after_key']
                if len(after) == 1:
                    result.update(after=after[quantity_name])
                else:
                    for key in after:
                        if key != quantity_name:
                            result.update(after=after[key])
                            break

            return result

        quantity_results = {
            quantity_name[9:]: create_quantity_result(quantity_name[9:], quantity)
            for quantity_name, quantity in aggs.items()
            if quantity_name.startswith('quantity:')
        }

        if len(quantity_results) > 0:
            result.update(quantities=quantity_results)

        return result

    def __str__(self):
        return json.dumps(self._search.to_dict(), indent=2)


def flat(obj, prefix=None):
    '''
    Helper that translates nested result objects into flattened dicts with
    ``domain.quantity`` as keys.
    '''
    if isinstance(obj, dict):
        result = {}
        for key, value in obj.items():
            if isinstance(value, dict):
                value = flat(value)
                for child_key, child_value in value.items():
                    result['%s.%s' % (key, child_key)] = child_value

            else:
                result[key] = value

        return result
    else:
        return obj


def _api_to_es_query(query: api_models.Query) -> Q:
    '''
    Creates an ES query based on the API's query model. This needs to be a normalized
    query expression with explicit objects for logical, set, and comparison operators.
    Shorthand notations ala ``quantity:operator`` are not supported here; this
    needs to be resolved via the respective pydantic validator. There is also no
    validation of quantities and types.
    '''
    def quantity_to_es(name: str, value: api_models.Value) -> Q:
        # TODO depends on keyword or not, value might need normalization, etc.
        quantity = search_quantities[name]
        return Q('match', **{quantity.search_field: value})

    def parameter_to_es(name: str, value: api_models.QueryParameterValue) -> Q:

        if isinstance(value, api_models.All):
            return Q('bool', must=[
                quantity_to_es(name, item)
                for item in value.op])

        if isinstance(value, api_models.Any_):
            return Q('bool', should=[
                quantity_to_es(name, item)
                for item in value.op])

        if isinstance(value, api_models.None_):
            return Q('bool', must_not=[
                quantity_to_es(name, item)
                for item in value.op])

        if isinstance(value, api_models.ComparisonOperator):
            quantity = search_quantities[name]
            return Q('range', **{quantity.search_field: {
                type(value).__name__.lower(): value.op}})

        # list of values is treated as an "all" over the items
        if isinstance(value, list):
            return Q('bool', must=[
                quantity_to_es(name, item)
                for item in value])

        return quantity_to_es(name, value)

    def query_to_es(query: api_models.Query) -> Q:
        if isinstance(query, api_models.LogicalOperator):
            if isinstance(query, api_models.And):
                return Q('bool', must=[query_to_es(operand) for operand in query.op])

            if isinstance(query, api_models.Or):
                return Q('bool', should=[query_to_es(operand) for operand in query.op])

            if isinstance(query, api_models.Not):
                return Q('bool', must_not=query_to_es(query.op))

            raise NotImplementedError()

        if not isinstance(query, dict):
            raise NotImplementedError()

        # dictionary is like an "and" of all items in the dict
        if len(query) == 0:
            return Q()

        if len(query) == 1:
            key = next(iter(query))
            return parameter_to_es(key, query[key])

        return Q('bool', must=[
            parameter_to_es(name, value) for name, value in query.items()])

    return query_to_es(query)


def _api_to_es_statistic(es_search: Search, name: str, statistic: Statistic) -> A:
    '''
    Creates an ES aggregation based on the API's statistic model.
    '''

    quantity = search_quantities[statistic.quantity.value]
    if quantity.statistic_values is not None:
        statistic.size = len(quantity.statistic_values)

    terms_kwargs = {}
    if statistic.value_filter is not None:
        terms_kwargs['include'] = '.*%s.*' % statistic.value_filter

    order_type = '_count' if statistic.order.type_ == AggregationOrderType.entries else '_key'
    statistic_agg = es_search.aggs.bucket('statistic:%s' % name, A(
        'terms',
        field=quantity.search_field,
        size=statistic.size,
        order={order_type: statistic.order.direction.value},
        **terms_kwargs))

    for metric in statistic.metrics:
        metric_quantity = metrics[metric.value]
        statistic_agg.metric('metric:%s' % metric_quantity.metric_name, A(
            metric_quantity.metric,
            field=metric_quantity.search_field))


def _es_to_api_statistics(es_response, name: str, statistic: Statistic) -> StatisticResponse:
    '''
    Creates a StatisticResponse from elasticsearch response on a request executed with
    the given statistics.
    '''
    quantity = search_quantities[statistic.quantity.value]

    es_statistic = es_response.aggs['statistic:' + name]
    statistic_data = {}
    for bucket in es_statistic.buckets:
        value_data = dict(entries=bucket.doc_count)
        for metric in statistic.metrics:
            value_data[metric.value] = bucket['metric:' + metric.value].value
        statistic_data[bucket.key] = value_data

    if quantity.statistic_values is not None:
        for value in quantity.statistic_values:
            if value not in statistic_data:
                statistic_data[value] = dict(entries=0, **{
                    metric.value: 0 for metric in statistic.metrics})

    return StatisticResponse(data=statistic_data, **statistic.dict(by_alias=True))


def _api_to_es_aggregation(es_search: Search, name: str, agg: Aggregation) -> A:
    '''
    Creates an ES aggregation based on the API's aggregation model.
    '''
    quantity = search_quantities[agg.quantity.value]
    terms = A('terms', field=quantity.search_field, order=agg.pagination.order.value)

    # We are using elastic searchs 'composite aggregations' here. We do not really
    # compose aggregations, but only those pseudo composites allow us to use the
    # 'after' feature that allows to scan through all aggregation values.
    order_by = agg.pagination.order_by
    if order_by is None:
        composite = dict(sources={name: terms}, size=agg.pagination.page_size)
    else:
        order_quantity = search_quantities[order_by]
        sort_terms = A('terms', field=order_quantity.search_field, order=agg.pagination.order.value)
        composite = dict(sources=[{order_by: sort_terms}, {quantity.name: terms}], size=agg.pagination.page_size)

    if agg.pagination.page_after_value is not None:
        if order_by is None:
            composite['after'] = {name: agg.pagination.page_after_value}
        else:
            order_value, quantity_value = agg.pagination.page_after_value.split(':')
            composite['after'] = {quantity.name: quantity_value, order_quantity.name: order_value}

    composite_agg = es_search.aggs.bucket('agg:%s' % name, 'composite', **composite)

    if agg.entries is not None and agg.entries.size > 0:
        kwargs: Dict[str, Any] = {}
        if agg.entries.required is not None:
            if agg.entries.required.include is not None:
                kwargs.update(_source=dict(includes=agg.entries.required.include))
            else:
                kwargs.update(_source=dict(excludes=agg.entries.required.exclude))

        composite_agg.metric('entries', A('top_hits', size=agg.entries.size, **kwargs))

    # additional cardinality to get total
    es_search.aggs.metric('agg:%s:total' % name, 'cardinality', field=quantity.search_field)


def _es_to_api_aggregation(es_response, name: str, agg: Aggregation) -> AggregationResponse:
    '''
    Creates a AggregationResponse from elasticsearch response on a request executed with
    the given aggregation.
    '''
    order_by = agg.pagination.order_by
    quantity = search_quantities[agg.quantity.value]
    es_agg = es_response.aggs['agg:' + name]

    def get_entries(agg):
        if 'entries' in agg:
            return [item['_source'] for item in agg.entries.hits.hits]
        else:
            return None

    if agg.pagination.order_by is None:
        agg_data = {
            bucket.key[name]: AggregationDataItem(size=bucket.doc_count, data=get_entries(bucket))
            for bucket in es_agg.buckets}
    else:
        agg_data = {
            bucket.key[quantity.search_field]: AggregationDataItem(size=bucket.doc_count, data=get_entries(bucket))
            for bucket in es_agg.buckets}

    aggregation_dict = agg.dict(by_alias=True)
    pagination = PaginationResponse(
        total=es_response.aggs['agg:%s:total' % name]['value'],
        **aggregation_dict.pop('pagination'))

    if 'after_key' in es_agg:
        after_key = es_agg['after_key']
        if order_by is None:
            pagination.next_page_after_value = after_key[name]
        else:
            str_values = [str(v) for v in after_key.to_dict().values()]
            pagination.next_page_after_value = ':'.join(str_values)

    return AggregationResponse(data=agg_data, pagination=pagination, **aggregation_dict)


def search(
        owner: str = 'public',
        query: Query = None,
        pagination: EntryPagination = None,
        required: MetadataRequired = None,
        aggregations: Dict[str, Aggregation] = {},
        statistics: Dict[str, Statistic] = {},
        user_id: str = None) -> SearchResponse:

    # The first half of this method creates the ES query. Then the query is run on ES.
    # The second half is about transforming the ES response to a SearchResponse.

    # query and owner
    if query is None:
        query = {}
    es_query = _api_to_es_query(query)
    if owner is not None:
        es_query &= _owner_es_query(owner=owner, user_id=user_id)

    # pagination
    if pagination is None:
        pagination = EntryPagination()

    search = Search(index=config.elastic.index_name)

    search = search.query(es_query)
    order_field = search_quantities[pagination.order_by].search_field
    sort = {order_field: pagination.order.value}
    if order_field != 'calc_id':
        sort['calc_id'] = pagination.order.value
    search = search.sort(sort)
    search = search.extra(size=pagination.page_size)
    if pagination.page_after_value:
        search = search.extra(search_after=pagination.page_after_value.rsplit(':', 1))

    if pagination.page:
        search = search[(pagination.page - 1) * pagination.page_size: pagination.page * pagination.page_size]

    # required
    if required:
        if required.include is not None and pagination.order_by not in required.include:
            required.include.append(pagination.order_by)
        if required.exclude is not None and pagination.order_by in required.exclude:
            required.exclude.remove(pagination.order_by)
        search = search.source(includes=required.include, excludes=required.exclude)

    # statistics
    for name, statistic in statistics.items():
        _api_to_es_statistic(search, name, statistic)

    # aggregations
    for name, agg in aggregations.items():
        _api_to_es_aggregation(search, name, agg)

    # execute
    try:
        es_response = search.execute()
    except RequestError as e:
        raise ElasticSearchError(e)

    more_response_data = {}

    # pagination
    next_page_after_value = None
    if 0 < len(es_response.hits) < es_response.hits.total:
        last = es_response.hits[-1]
        if order_field == 'calc_id':
            next_page_after_value = last['calc_id']
        else:
            after_value = last
            for order_field_segment in order_field.split('.'):
                after_value = after_value[order_field_segment]
            next_page_after_value = '%s:%s' % (after_value, last['calc_id'])
    pagination_response = PaginationResponse(
        total=es_response.hits.total,
        next_page_after_value=next_page_after_value,
        **pagination.dict())

    # statistics
    if len(statistics) > 0:
        more_response_data['statistics'] = cast(Dict[str, Any], {
            name: _es_to_api_statistics(es_response, name, statistic)
            for name, statistic in statistics.items()})

    # aggregations
    if len(aggregations) > 0:
        more_response_data['aggregations'] = cast(Dict[str, Any], {
            name: _es_to_api_aggregation(es_response, name, aggregation)
            for name, aggregation in aggregations.items()})

    more_response_data['es_query'] = es_query.to_dict()

    return SearchResponse(
        owner=owner,
        query=query,
        pagination=pagination_response,
        required=required,
        data=[_es_to_entry_dict(hit, required) for hit in es_response.hits],
        **more_response_data)


def update_by_query(update_script: str, owner: str = 'public', query: Query = None, user_id: str = None, **kwargs):
    '''
    Uses the given painless script to update the entries by given query.

    In most cases, the elasticsearch entry index should not be updated field by field;
    you should run `index_all` instead and fully replace documents from mongodb and
    archive files.

    This method provides a faster direct method to update individual fiels, e.g. to quickly
    update fields for editing operations.
    '''

    if query is None:
        query = {}
    es_query = _api_to_es_query(query)
    es_query &= _owner_es_query(owner=owner, user_id=user_id)

    body = {
        'script': {
            'source': update_script,
            'lang': 'painless'
        },
        'query': es_query.to_dict()
    }

    body['script'].update(**kwargs)

    try:
        result = infrastructure.elastic_client.update_by_query(
            body=body, index=config.elastic.index_name)
    except TransportError as e:
        utils.get_logger(__name__).error(
            'es update_by_query script error', exc_info=e,
            es_info=json.dumps(e.info, indent=2))
        raise e

    return result



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from siestaparser.metainfo import siesta_autogenerated
from nomad.datamodel.metainfo import public

m_package = Package(
    name='siesta_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='siesta.nomadmetainfo.json'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_siesta_system_label = Quantity(
        type=str,
        shape=[],
        description='''
        system label specified in siesta calculation.  This determines many of the
        filenames.
        ''',
        a_legacy=LegacyDefinition(name='x_siesta_system_label'))

    x_siesta_system_name = Quantity(
        type=str,
        shape=[],
        description='''
        system name specified in siesta calculation.  This is purely a description.
        ''',
        a_legacy=LegacyDefinition(name='x_siesta_system_name'))

    x_siesta_arch = Quantity(
        type=str,
        shape=[],
        description='''
        architecture
        ''',
        a_legacy=LegacyDefinition(name='x_siesta_arch'))

    x_siesta_compilerflags = Quantity(
        type=str,
        shape=[],
        description='''
        basis set type
        ''',
        a_legacy=LegacyDefinition(name='x_siesta_compilerflags'))


class x_siesta_section_input(siesta_autogenerated.x_siesta_section_input):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='x_siesta_section_input'))

    x_siesta_xc_authors = Quantity(
        type=str,
        shape=[],
        description='''
        XC authors (really: XC functional) in siesta calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_siesta_xc_authors'))

    x_siesta_typeofrun = Quantity(
        type=str,
        shape=[],
        description='''
        type of run as categorized by Siesta (optimization/md etc.)
        ''',
        a_legacy=LegacyDefinition(name='x_siesta_typeofrun'))


class section_basis_set(public.section_basis_set):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_basis_set'))

    x_siesta_pao_basis_type = Quantity(
        type=str,
        shape=[],
        description='''
        basis set type
        ''',
        a_legacy=LegacyDefinition(name='x_siesta_pao_basis_type'))


m_package.__init_metainfo__()



import matplotlib.pyplot as plt
from bitarray import bitarray

def plot_FP_in_grid(byte_fingerprint, grid, show = True, label = '', axes = None, **kwargs):
    x=[]
    y=[]
    all_width=[]
    bin_fp=bitarray()
    bin_fp.frombytes(bytes.fromhex(byte_fingerprint.bins))
    grid_indices=byte_fingerprint.indices
    plotgrid=grid.grid()
    plotgrid=plotgrid[grid_indices[0]:grid_indices[1]]
    bit_position=0
    for index,item in enumerate(plotgrid):
        if index<len(plotgrid)-1:
            width=plotgrid[index+1][0]-item[0]
        else:
            width=abs(item[0]-plotgrid[index-1][0])
        for idx, dos_value in enumerate(item[1]):
            if bin_fp[bit_position]==1 and (idx == len(item[1])-1 or bin_fp[bit_position+1] == 0 ):
                x.append(item[0])
                y.append(dos_value)
                all_width.append(width)
            bit_position+=1
    if axes == None:
        plt.bar(x,y,width=all_width,align='edge', label = label, **kwargs)
    else:
        axes.bar(x,y,width=all_width,align='edge', label = label, **kwargs)
    if show:
        plt.show()
